# -*- coding: utf-8 -*-
'''
ReLib Backends
'''
import logging
import os
import six
from six.moves import cStringIO as StringIO
import zipfile

from .relib import Backend, BackendFileNotFound, register

log = logging.getLogger(__name__)


@register
class S3Backend(Backend):
    """
    S3 backend for TableStore.
    """

    __scheme__ = 's3'
    default_format = 'pickle'

    def __init__(self, bucket_name, folder_name, region_name=None, etag=None):
        import boto3
        self.s3_client = boto3.client('s3', region_name=region_name)
        self.bucket_name = bucket_name
        self.folder_name = folder_name.lstrip('/')  # Strip leading slashes
        self.region_name = region_name
        self.etag = etag

    @classmethod
    def create_from_url_parts(cls, parts, query):
        if 'region' in query:
            region_name = query['region'][0]
        else:
            region_name = None
        return cls(bucket_name=parts.hostname, folder_name=parts.path, region_name=region_name)

    def get_url(self):
        url = 's3://{}/{}'.format(self.bucket_name, self.folder_name)
        if self.region_name:
            url += '?region=' + self.region_name
        return url

    def __str__(self):
        return "S3Backend'{}'".format(self.get_url())

    def get_key_name(self, file_name):
        return '{}/{}'.format(self.folder_name, file_name)

    def save_data(self, file_name, data):
        return self._save_data_with_bucket_logic(file_name, data, try_create_bucket=True)

    def _save_data_with_bucket_logic(self, file_name, data, try_create_bucket):
        from botocore.client import ClientError
        f = six.BytesIO(data)
        key_name = self.get_key_name(file_name)
        log.debug("Uploading %s bytes to s3://%s/%s", len(data), self.bucket_name, key_name)
        try:
            self.s3_client.upload_fileobj(
                f,
                self.bucket_name,
                key_name,
                ExtraArgs={'ContentType': 'application/json'},
            )
        except ClientError as e:
            if 'NoSuchBucket' in str(e) and try_create_bucket:
                self.s3_client.create_bucket(
                    Bucket=self.bucket_name,
                    CreateBucketConfiguration={'LocationConstraint': self.region_name}
                )
                return self._save_data_with_bucket_logic(file_name, data, try_create_bucket=False)
            else:
                raise

    def load_data(self, file_name):
        from botocore.client import ClientError
        key_name = self.get_key_name(file_name)
        log.debug("Downloading s3://%s/%s", self.bucket_name, key_name)
        f = six.BytesIO()
        try:
            self.s3_client.download_fileobj(self.bucket_name, key_name, f)
        except ClientError as e:
            if '404' in str(e):
                raise BackendFileNotFound
            raise
        return f.getvalue()


@register
class RedisBackend(Backend):

    __scheme__ = 'redis'
    default_format = 'pickle'
    is_cache = True

    def __init__(self, host=None, port=None, db=None, prefix=None, expire_sec=None):
        import redis
        host = host or 'localhost'
        port = port or 6379
        db = db or 0
        self.prefix = prefix or ''
        self.expire_sec = expire_sec

        self.conn = redis.StrictRedis(
            host=host,
            port=port,
            db=db,
            socket_timeout=5.0,
        )

        self.host, self.port, self.db = host, port, db
        log.debug("%s initialized.", self)

    @classmethod
    def create_from_url_parts(cls, parts, query):
        db = int(query['db'][0]) if 'db' in query else None
        prefix = query['prefix'][0] if 'prefix' in query else ''
        expire_sec = int(query['expire_sec'][0]) if 'expire_sec' in query else None
        b = cls(host=parts.hostname, port=parts.port, db=db, prefix=prefix, expire_sec=expire_sec)
        return b  # ZipEncoded(b)

    @classmethod
    def create_from_server_info(cls, host, port, domain_name):
        b = cls(
            host=host,
            port=port,
            prefix=domain_name,
            expire_sec=None,  # Never expires
            )
        return b  # ZipEncoded(b)

    def __str__(self):
        return "RedisBackend'{}:{}#{}, prefix={}'".format(self.host, self.port, self.db, self.prefix)

    def get_key_name(self, file_name):
        return 'relib:drift-config:{}:{}'.format(self.prefix, file_name)

    def save_data(self, file_name, data):
        # data is bytes.  redis will do no encoding
        key_name = self.get_key_name(file_name)
        log.debug("Adding %s bytes to Redis:%s with expiry:%s", len(data), key_name, self.expire_sec)
        self.conn.set(key_name, data)
        if self.expire_sec is not None:
            self.conn.expire(key_name, self.expire_sec)
        self.conn.set

    def load_data(self, file_name):
        # since the client is not created with "decode_responses", the result here is bytes
        key_name = self.get_key_name(file_name)
        log.debug("Reading from Redis:%s", key_name)
        data = self.conn.get(key_name)
        if data is None:
            log.warning("Redis cache doesn't have '{}'. (Is it expired?)".format(key_name))
            raise BackendFileNotFound
        return data

    def get_url(self):
        return "redis://{}:{}/{}?prefix={}".format(self.host, self.port, self.db, self.prefix)


@register
class FileBackend(Backend):

    __scheme__ = 'file'

    def __init__(self, folder_name):
        if '~' in folder_name:
            # Expand user and trim whatever was in front of the ~ char.
            folder_name = os.path.expanduser('~') + folder_name.split('~', 1)[1]

        folder_name = folder_name.replace('/', os.sep)  # Adjust to Windows platform mainly

        if not os.path.exists(folder_name):
            os.makedirs(folder_name)
        self.folder_name = folder_name

    @classmethod
    def create_from_url_parts(cls, parts, query):
        # combine host and path into one
        path = parts.netloc or ''  # Change None to '' if needed.
        path += parts.path
        return cls(folder_name=path)

    def get_url(self):
        path = self.folder_name
        path = path.replace(os.path.expanduser('~'), '~')
        path = path.replace('\\', '/')  # De-Windowize, if needed
        return path

    def __str__(self):
        return "FileBackend'{}'".format(self.folder_name)

    def get_filename(self, file_name):
        file_name = file_name.replace('/', os.sep)  # Adjust to Windows platform mainly
        return os.path.join(self.folder_name, file_name)

    # save and load bytes data
    def save_data(self, file_name, data):
        path_name = self.get_filename(file_name)

        # Create subdirs if neccessary
        dir_name = os.path.dirname(path_name)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        with open(path_name, 'wb') as f:
            log.debug("Writing %s bytes to %s", len(data), path_name)
            f.write(data)

    def load_data(self, file_name):
        path_name = self.get_filename(file_name)
        log.debug("Reading from %s", path_name)

        import os.path
        if not os.path.exists(path_name):
            raise BackendFileNotFound

        with open(path_name, 'rb') as f:
            return f.read()


@register
class MemoryBackend(Backend):

    __scheme__ = 'memory'
    archive = {}

    def __init__(self, folder_name):
        self.folder_name = folder_name
        MemoryBackend.archive[folder_name] = {}

    def __del__(self):
        del MemoryBackend.archive[self.folder_name]

    @classmethod
    def create_from_url_parts(cls, parts, query):
        return cls(folder_name=parts.path)

    def get_url(self):
        return 'memory://' + self.folder_name

    def __str__(self):
        return "MemoryBackend'{}'".format(self.folder_name)

    # save and load whatever data is provided.  It shoud be bytes
    def save_data(self, file_name, data):
        MemoryBackend.archive[self.folder_name][file_name] = data

    def load_data(self, file_name):
        if file_name not in MemoryBackend.archive[self.folder_name]:
            raise BackendFileNotFound
        return MemoryBackend.archive[self.folder_name][file_name]


class ZipEncoded(Backend):
    """Aggregate class which serializes to and from a single zip file."""
    def __init__(self, aggregate):
        self.aggregate = aggregate

    def start_saving(self):
        self._fp = StringIO()
        self._zipfile = zipfile.ZipFile(self._fp, mode='w', compression=zipfile.ZIP_DEFLATED)

    def done_saving(self):
        self.aggregate.save_data("_zipped.zip", self._fp.getvalue())

    def start_loading(self):
        self._fp = StringIO()
        self._fp.write(self.aggregate.load_data("_zipped.zip"))
        self._fp.seek(0)
        self._zipfile = zipfile.ZipFile(self._fp)

    def done_loading(self):
        pass

    def save_data(self, file_name, data):
        self._zipfile.writestr(file_name, data)

    def load_data(self, file_name):
        self._zipfile.read(file_name)



# -*- coding: utf-8 -*-
'''
Drift Config Utility and helpers functions

some important bits to do:

tier init eu-west/..../blah


concepts:

organization
tier
tenant
service/deployable



objects:

deployable + name,ssh_key = used by several commands (ami, quickdeploy, remotecmd, ssh, logs, tier)

deployable + name,api               = used by api-router configgenerator, kaleo web
deployable + autoscaling,release    = used by api-router configgenerator




global map rules for table files:
**********************************

foreign_key='organizations.organization_name'
template='organizations/{}'

ex:

table 'organizations'
organizations.json  # note, unchanged as the key field is pk, not fk.


table 'products'
organizations/nova/products.json
organizations/1939games/products.json

table 'tenant-names'
organizations/nova/tenant-names.json
organizations/1939games/tenant-names.json


table 'tenants'
organizations/nova/tenants/tenants.DEVNORTH.nova-live.json
organizations/1939games/tenants/tenants.DEVNORTH.1939-kards.json







drift-config core tables:


domain (single row):
    domain_name             string, required
    display_name            string
    origin                  string, required


organizations:
    organization_name       string, pk
    short_name              string, required
    display_name:           string
    state                   enum initializing|active|disabled|deleted, default=active


tiers:
    tier_name               string, pk
    is_live                 boolean, default=true
    state                   enum initializing|active|disabled|deleted, default=active


deployable-names:
    deployable_name         string, pk
    display_name            string, required


deployables:
    tier_name               string, pk, fk->tiers
    deployable_name         string, pk, fk->deployables
    version                 string
    is_active               boolean, default=false


products:
    product_name            string, pk
    organization_name       string, fk->organizations, required
    state                   enum initializing|active|disabled|deleted, default=active
    deployables             array of strings, required


tenant-names:
    # a tenant name is unique across all tiers
    tenant_name             string, pk
    product_name            string, fk->products, required
    organization_name       string, fk->organizations, required
    tier_name               string, pk, fk->tiers
    reserved_at             date-time, default=@@utcnow
    reserved_by             string


tenants:
    tier_name               string, pk, fk->tiers
    deployable_name         string, pk, fk->deployables
    tenant_name             string, pk, fk->tenants
    state                   enum initializing|active|disabled|deleted, default=initializing

    meta
        use_subfolder=true
        group_by=tier_name,tenant_name


public-keys:
    tier_name               string, pk, fk->tiers
    deployable_name         string, pk, fk->deployables
    keys                    array

    meta
        subfolder_name=authentication


platforms:
    tier_name               string, pk, fk->tiers
    deployable_name         string, pk, fk->deployables
    tenant_name             string, pk, fk->tenants
    providers               array

    meta
        subfolder_name=authentication

users:
    organization_name   string, pk, fk->organizations, required
    user_name           string, pk, required
    create_date         datetime, required, default=@@utcnow
    valid_until         datetime
    is_active           boolean, required, default=true
    password            string
    access_key          string
    is_service          boolean, required, default=false
    is_role_admin       boolean, required, default=false

    meta
        subfolder_name=authentication


/////////////////////////////////////////////////////////////////////////////////
// APP SPECIFIC CONFIG - api-router
/////////////////////////////////////////////////////////////////////////////////

api-router/

routing:
    tier_name               string, pk, fk->tiers
    deployable_name         string, pk, fk->deployables
    api                     string, required
    autoscaling             dict
    release_version         string


api-keys:
    api_key_name            string, pk
    product_name            string, fk->products, required
    in_use                  boolean, default=true, required
    create_date             datetime, default=@@utcnow
    key_type                enum product|custom, default=product, required
    custom_data             string


api-key-rules:
    product_name            string, pk, fk->products, required
    rule_name               string, pk

    assignment_order        integer, required
    version_patterns        array of strings, required
    is_active               boolean, required, default=true

    rule_type               enum pass|redirect|reject, required, default=pass
    response_header         dict
    redirect:
        tenant_name             string
    reject:
        status_code             integer
        response_body           dict


/////////////////////////////////////////////////////////////////////////////////
// APP SPECIFIC CONFIG - ue4-gameserver
/////////////////////////////////////////////////////////////////////////////////


ue4-gameservers/

ue4-gameservers-config (single row):
    build_archive_defaults
        region                  string
        bucket_name             string
        ue4_builds_folder       string


ue4-builds
    product_name            string, pk, fk->products, required

    s3_region               string
    bucket_name             string
    path                    string


gameservers-machines:
    product_name            string, pk, fk->products, required
    group_name              string, pk
    region                  string, pk
    platform                enum windows|linux, required
    autoscaling
        min                 integer
        max                 integer
        desired             integer
        instance_type       string, required


gameserver-instances:
    product_name            string, pk, fk->products, required
    group_name              string, pk, fk->gameservers-machines, required
    region                  string, pk, fk->gameservers-machines, required
    tenant_name             string, pk, fk->tenants, required
    ref:                    string, pk, required
    processes_per_machine   integer, required


/////////////////////////////////////////////////////////////////////////////////
// APP SPECIFIC CONFIG - drift-base
/////////////////////////////////////////////////////////////////////////////////

drift-base/

client-versions:
    product_name            string fk->products
    release_version         string



******************************************************************

authentication:
    tier                    fk->tiers
    deployable_name         fk->deployables
    keys
        issued              datetime
        expires             datetime
        public_key          string
        private_key         string      (note, this entry only accessible to the issuing service)

    authentications
        # 3rd party authentication
        tenant_name         fk->tenants
        provider_name       string
        provider_details    dict

users:
    # replacing the global 'service_user'
    organization_name   string, pk, fk->organizations, required
    user_name           string, pk, required
    create_date         datetime, required, default=@@utcnow
    valid_until         datetime
    is_active           boolean, required, default=true
    password            string
    access_key          string
    is_service          boolean, required, default=false
    is_role_admin       boolean, required, default=false


user-acl-roles:
    organization_name   string, pk, fk->organizations, required
    user_name           string, pk, fk->users, required
    role_name           string, pk, fk->user-roles, required
    tenant_name         string, fk->tenants


# dynamically populated by deployables during "init" phase
access-roles:
    role_name               string, pk
    deployable_name         string, fk->deployables
    description             string


use case:

tenant: superkaiju
login: directivegames.matti
pass:  bobo

-> lookup in user-acl:
    no record -> if 'superuser' in  org.role:
        assign ['admin'] to roles



application specific tables:


aws_tools:
    services
        deployable_name     fk->deployables
        ssh_key             string


drift_base:
    tier                    fk->tiers
    db_connection_info      dict (RDS stuff)


server_daemon:
    access_key              string
    secret_key              string




refactor "deployable":

drift-config defines:

    service/deployable:
        name:       name of deployable
        tenants:    array of tenant names.


api-router specifics:
        "deployables":
        {
            "name": "<deployable name>",
            "api": "<api shortname>",
            "autoscaling": {},
            "release": "..."
        },
        "api_keys":
        [
            {},...
        ]


aws commands extend service/deployable:

        "aws_tools":
        {
            "ssh_key": "keyname"
        }




'''
import logging
from datetime import datetime

from driftconfig.relib import TableStore, create_backend
import driftconfig.relib
from driftconfig.util import get_default_drift_config_and_source
from driftconfig.backends import RedisBackend

log = logging.getLogger(__name__)


# tenant name:
"<org name>-<tier name>-<product name>"
'''
directivegames-superkaiju-LIVENORTH.dg-api.com
directivegames-superkaiju-DEVNORTH.dg-api.com

superkaiju.dg-api.com
directivegames-borkbork

.dg-api.com


'''


def get_drift_table_store():
    """
    Create a Data Store which contains all Core Drift tables.
    """

    # RULE: pk='tier_name'='LIVENORTH', role['liveops', 'admin', 'service']

    ts = TableStore()

    domain = ts.add_table('domain', single_row=True)
    domain.add_schema({
        'type': 'object',
        'properties': {
            'domain_name': {'type': 'string'},
            'display_name': {'type': 'string'},
            'origin': {'type': 'string'},
        },
        'required': ['domain_name', 'origin'],
    })
    domain.add_default_values({'domain_name': '', 'origin': ''})

    organizations = ts.add_table('organizations')
    organizations.add_primary_key('organization_name')
    organizations.add_unique_constraint('short_name')
    organizations.add_schema({
        'type': 'object',
        'properties': {
            'organization_name': {'pattern': r'^([a-z0-9]){2,20}$'},
            'short_name': {'pattern': r'^([a-z0-9]){2,20}$'},
            'display_name': {'type': 'string'},
            'state': {'enum': ['initializing', 'active', 'disabled', 'deleted']},
        },
        'required': ['short_name'],
    })
    organizations.add_default_values({'state': 'active'})

    tiers = ts.add_table('tiers')
    tiers.add_primary_key('tier_name')
    tiers.add_schema({
        'type': 'object',
        'properties': {
            'tier_name': {'pattern': r'^([A-Z]){3,20}$'},
            'is_live': {'type': 'boolean'},
            'state': {'enum': ['initializing', 'active', 'disabled', 'deleted']},
        },
        'required': ['is_live'],
    })
    tiers.add_default_values({'is_live': True, 'state': 'active'})

    deployable_names = ts.add_table('deployable-names')
    deployable_names.add_primary_key('deployable_name')
    deployable_names.add_schema({
        'type': 'object',
        'properties': {
            'deployable_name': {'pattern': r'^[a-z]([a-z-0-9]){2,20}$'},
            'display_name': {'type': 'string'},
            'resources': {'type': 'array', 'item': 'string'},
            'resource_attributes': {'type': 'object'},
        },
        'required': ['display_name', 'resources'],
    })
    deployable_names.add_default_values({'resources': []})

    deployables = ts.add_table('deployables')
    deployables.add_primary_key('tier_name,deployable_name')
    deployables.add_foreign_key('tier_name', 'tiers')
    deployables.add_foreign_key('deployable_name', 'deployable-names')
    deployables.add_schema({
        'type': 'object',
        'properties': {
            'release': {'type': 'string'},
            'is_active': {'type': 'boolean'},
            'reason_inactive': {'type': 'string'},
        },
        'required': ['is_active'],
    })
    deployables.add_default_values({'is_active': True})

    products = ts.add_table('products')
    products.add_primary_key('product_name')
    products.add_foreign_key('organization_name', 'organizations')
    products.add_schema({
        'type': 'object',
        'properties': {
            'product_name': {'pattern': r'^([a-z0-9-]){3,35}$'},
            'state': {'enum': ['initializing', 'active', 'disabled', 'deleted']},
            'deployables': {'type': 'array', 'items': {'type': 'string'}},
        },
        'required': ['organization_name', 'deployables'],
    })
    products.add_default_values({'state': 'active', 'deployables': []})

    # Waiting a little bit with fixups of tenant-names and tenants table schema
    if "temporary fixy fix":
        tenant_names = ts.add_table('tenant-names')
        tenant_names.add_primary_key('tenant_name')
        tenant_names.add_foreign_key('product_name', 'products')
        tenant_names.add_foreign_key('organization_name', 'organizations')
        tenant_names.add_foreign_key('tier_name', 'tiers')
        tenant_names.add_unique_constraint('alias')
        tenant_names.add_schema({
            'type': 'object',
            'properties': {
                'tenant_name': {'pattern': r'^([a-z0-9-]){3,30}$'},
                'alias': {'pattern': r'^([a-z0-9-]){3,30}$'},
                'reserved_at': {'format': 'date-time'},
                'reserved_by': {'type': 'string'},
            },
            'required': ['product_name', 'organization_name', 'tier_name'],
        })
        tenant_names.add_default_values({'reserved_at': '@@utcnow'})

        tenants = ts.add_table('tenants')
        tenants.add_primary_key('tier_name,deployable_name,tenant_name')
        tenants.set_row_as_file(subfolder_name=tenants.name, group_by='tier_name,tenant_name')
        tenants.add_foreign_key('tier_name', 'tiers')
        tenants.add_foreign_key('deployable_name', 'deployable-names')
        tenants.add_foreign_key('tenant_name', 'tenant-names')
        tenants.add_schema({
            'type': 'object',
            'properties': {
                'state': {'enum': [
                    'initializing', 'active', 'disabled', 'uninitializing', 'deleted'
                ]},
            },
        })
        tenants.add_default_values({'state': 'initializing'})
    else:
        tenant_names = ts.add_table('tenant-names')
        tenant_names.add_primary_key('tenant_name,product_name')
        tenant_names.set_row_as_file(subfolder_name='tenants', group_by='product_name')
        tenant_names.add_unique_constraint('tenant_name')
        tenant_names.add_foreign_key('product_name', 'products')
        tenant_names.add_foreign_key('organization_name', 'organizations')
        tenant_names.add_foreign_key('tier_name', 'tiers')
        tenant_names.add_schema({
            'type': 'object',
            'properties': {
                'tenant_name': {'pattern': r'^([a-z0-9-]){3,30}$'},
                'reserved_at': {'format': 'date-time'},
                'reserved_by': {'type': 'string'},
            },
            'required': ['product_name', 'organization_name', 'tier_name'],
        })
        tenant_names.add_default_values({'reserved_at': '@@utcnow'})

        tenants = ts.add_table('tenants')
        tenants.add_primary_key('deployable_name,tenant_name')
        tenants.set_row_as_file(subfolder_name='tenants', group_by='tenant_name')
        tenants.add_foreign_key('deployable_name', 'deployable-names')
        tenants.add_foreign_key('tenant_name', 'tenant-names')
        tenants.add_schema({
            'type': 'object',
            'properties': {
                'state': {'enum': [
                    'initializing',  # Tenant resources are being provisioned
                    'active',  # Tenant resources are online and accessible
                    'disabled',  # Tenant resources are offline and not accessible
                    'uninitializing',  # Tenant resources are being unprovisoned/deleted.
                    'deleted',  # Tenant resources have been deleted.
                ]},
            },
        })
        tenants.add_default_values({'state': 'initializing'})

    public_keys = ts.add_table('public-keys')
    public_keys.set_row_as_file(subfolder_name='authentication')
    public_keys.add_primary_key('tier_name,deployable_name')
    public_keys.add_foreign_key('tier_name', 'tiers')
    public_keys.add_foreign_key('deployable_name', 'deployable-names')
    public_keys.add_schema({
        'type': 'object',
        'properties': {
            'keys': {'type': 'array', 'items': {
                'type': 'object',
                'properties': {
                    'issued': {'format': 'date-time'},
                    'expires': {'format': 'date-time'},
                    'public_key': {'type': 'string'},
                    'private_key': {'type': 'string'},
                },
            }},
        },
    })

    platforms = ts.add_table('platforms')
    platforms.set_row_as_file(subfolder_name='authentication')
    platforms.add_primary_key('product_name,provider_name')
    platforms.add_foreign_key('product_name', 'products')
    platforms.add_schema({
        'type': 'object',
        'properties': {
            'provider_details': {'type': 'object'},
        },
        'required': ['provider_details'],
    })

    '''
    users:
        organization_name   string, pk, fk->organizations, required
        user_name           string, pk, required
        create_date         datetime, required, default=@@utcnow
        valid_until         datetime
        is_active           boolean, required, default=true
        password            string
        access_key          string
        is_service          boolean, required, default=false
        is_role_admin       boolean, required, default=false

        meta
            subfolder_name=authentication
    '''
    users = ts.add_table('users')
    users.set_row_as_file(subfolder_name='authentication')
    users.add_primary_key('organization_name,user_name')
    users.add_foreign_key('organization_name', 'organizations')
    users.add_schema({
        'type': 'object',
        'properties': {
            'user_name': {'pattern': r'^([a-z0-9_]){2,30}$'},
            'create_date': {'format': 'date-time'},
            'valid_until': {'format': 'date-time'},
            'is_active': {'type': 'boolean'},
            'password': {'type': 'string'},
            'access_key': {'type': 'string'},
            'is_service': {'type': 'boolean'},
            'is_role_admin': {'type': 'boolean'},
        },
        'required': ['create_date', 'is_active', 'is_service', 'is_role_admin'],
    })
    users.add_default_values({
        'create_date': '@@utcnow', 'is_active': True,
        'is_service': False, 'is_role_admin': False
    })

    '''
    # dynamically populated by deployables during "init" phase
    access-roles:
        role_name               string, pk
        deployable_name         string, fk->deployables, required
        description             string
    '''
    access_roles = ts.add_table('access-roles')
    access_roles.set_row_as_file(subfolder_name='authentication')
    access_roles.add_primary_key('role_name')
    platforms.add_foreign_key('deployable_name', 'deployable-names')
    access_roles.add_schema({
        'type': 'object',
        'properties': {
            'description': {'type': 'string'},
        },
        'required': ['deployable_name'],
    })

    '''
    users-acl:
        organization_name   string, pk, fk->organizations, required
        user_name           string, pk, fk->users, required
        role_name           string, pk, fk->user-roles, required
        tenant_name         string, fk->tenants
    '''
    users_acl = ts.add_table('users-acl')
    users_acl.set_row_as_file(subfolder_name='authentication')
    users_acl.add_primary_key('organization_name,user_name,role_name')
    users_acl.add_foreign_key('organization_name', 'organizations')
    users_acl.add_foreign_key('user_name', 'users')
    users_acl.add_foreign_key('role_name', 'access-roles')
    users_acl.add_foreign_key('tenant_name', 'tenant-names')

    # RELEASE MANAGEMENT - THIS SHOULDN'T REALLY BE IN THIS FILE HERE, or what?
    '''
    instances:
        tier_name               string, pk, fk->tiers
        deployable_name         string, pk, fk->deployables
        autoscaling
            min                 integer
            max                 integer
            desired             integer
            instance_type       string, required
        release_version         string
    '''
    instances = ts.add_table('instances')
    instances.set_subfolder_name('release-mgmt')
    instances.add_primary_key('tier_name,deployable_name')
    instances.add_foreign_key('tier_name', 'tiers')
    instances.add_foreign_key('deployable_name', 'deployable-names')
    instances.add_schema({'type': 'object', 'properties': {
        'autoscaling': {'type': 'object', 'properties': {
            'min': {'type': 'integer'},
            'max': {'type': 'integer'},
            'desired': {'type': 'integer'},
            'instance_type': {'type': 'string'},
        }},
        'release_version': {'type': 'string'},
    }})

    # API ROUTER STUFF - THIS SHOULDN'T REALLY BE IN THIS FILE HERE
    '''
    nginx:
        tier_name               string, pk, fk->tiers
    '''
    nginx = ts.add_table('nginx')
    nginx.add_primary_key('tier_name')
    nginx.set_subfolder_name('api-router',)
    nginx.add_schema({'type': 'object', 'properties': {
        'worker_rlimit_nofile': {'type': 'integer'},
        'worker_connections': {'type': 'integer'},
        'api_key_passthrough':
        {
            'type': 'array',
            'items':
            {
                'type': 'object',
                'properties':
                {
                    'key_name': {'type': 'string'},
                    'key_value': {'type': 'string'},
                    'product_name': {'type': 'string'},
                },
                'required': ['key_name', 'key_value', 'product_name'],
            }
        },
    }})

    ''' a yaml representation would be:
    ---
    type: object
    properties:
      worker_connections: {type: integer}
      worker_rlimit_nofile: {type: integer}
      api_key_passthrough:
        type: array
        items:
          type: object
          required: [key_name, key_value ,product_name]
          properties:
            key_name: {type: string}
            key_value: {type: string}
            product_name: {type: string}

    '''

    '''
    routing:
        deployable_name         string, pk, fk->deployables
        api                     string, required
    '''
    routing = ts.add_table('routing')
    routing.set_subfolder_name('api-router')
    routing.add_primary_key('deployable_name')
    routing.add_foreign_key('deployable_name', 'deployable-names')
    routing.add_schema({
        'type': 'object',
        'properties':
        {
            'api': {'type': 'string'},
            'requires_api_key': {'type': 'boolean'},
        },
        'required': ['requires_api_key'],
    })
    routing.add_default_values({'requires_api_key': True})

    '''
    api-keys:
        api_key_name            string, pk
        product_name            string, fk->products, required
        in_use                  boolean, default=true, required
        create_date             datetime, default=@@utcnow
        key_type                enum product|custom, default=product, required
        custom_data             string
    '''
    keys = ts.add_table('api-keys')
    keys.set_subfolder_name('api-router')
    keys.add_primary_key('api_key_name')
    keys.add_foreign_key('product_name', 'products')
    keys.add_schema({
        'type': 'object',
        'properties':
        {
            'in_use': {'type': 'boolean'},
            'create_date': {'format': 'date-time'},
            'key_type': {'enum': ['product', 'custom']},
            'custom_data': {'type': 'string'},
        },
        'required': ['in_use', 'key_type'],
    })
    keys.add_default_values({'in_use': True, 'create_date': '@@utcnow', 'key_type': 'product'})

    '''
    api-key-rules:
        product_name            string, pk, fk->products
        rule_name               string, pk

        assignment_order        integer, required
        version_patterns        array of strings, required
        is_active               boolean, required, default=true

        rule_type               enum pass|redirect|reject, required, default=pass
        response_header         dict
        redirect:
            tenant_name             string
        reject:
            status_code             integer
            response_body           dict
    '''
    keyrules = ts.add_table('api-key-rules')
    keyrules.set_subfolder_name('api-router')
    keyrules.add_primary_key('product_name,rule_name')
    keyrules.add_foreign_key('product_name', 'products')
    keyrules.add_schema({
        'type': 'object',
        'properties':
        {
            'assignment_order': {'type': 'integer'},
            'version_patterns': {'type': 'array', 'items': {'type': 'string'}},
            'is_active': {'type': 'boolean'},

            'rule_type': {'enum': ['pass', 'redirect', 'reject']},

            'response_header': {'type': 'object'},
            'redirect': {'type': 'object', 'properties': {
                'tenant_name': {'type': 'string'},
            }},
            'reject': {'type': 'object', 'properties': {
                'status_code': {'type': 'integer'},
                'response_body': {'type': 'object'},
            }},
        },
        'required': ['assignment_order', 'version_patterns', 'is_active', 'rule_type'],
    })
    keyrules.add_default_values({'is_active': True, 'rule_type': 'pass'})

    '''
    ue4-gameservers/

    ue4-gameservers-config (single row):
        build_archive_defaults
            region                  string
            bucket_name             string
            ue4_builds_folder       string
    '''
    ue4_gameservers_config = ts.add_table('ue4-gameservers-config', single_row=True)
    ue4_gameservers_config.set_subfolder_name('ue4-gameservers')
    ue4_gameservers_config.add_schema({
        'type': 'object',
        'properties':
        {
            'build_archive_defaults': {
                'type': 'object',
                'properties':
                {
                    'region': {'type': 'string'},
                    'bucket_name': {'type': 'string'},
                    'ue4_builds_folder': {'type': 'string'},
                },
            },
        },
    })

    '''
    ue4-build-artifacts
        product_name            string, pk, fk->products, required

        s3_region               string
        bucket_name             string
        path                    string
        command_line            string
    '''
    ue4_build_artifacts = ts.add_table('ue4-build-artifacts')
    ue4_build_artifacts.set_subfolder_name('ue4-gameservers')
    ue4_build_artifacts.add_primary_key('product_name')
    ue4_build_artifacts.add_foreign_key('product_name', 'products')
    ue4_build_artifacts.add_schema({
        'type': 'object',
        'properties':
        {
            's3_region': {'type': 'string'},
            'bucket_name': {'type': 'string'},
            'path': {'type': 'string'},
            'command_line': {'type': 'string'},
        },
        'required': ['s3_region', 'bucket_name', 'path', 'command_line'],
    })

    '''
    gameservers-machines:
        product_name            string, pk, fk->products, required
        group_name              string, pk
        region                  string, pk
        platform                enum windows|linux, required
        autoscaling
            min                 integer
            max                 integer
            desired             integer
            instance_type       string, required
    '''
    gameservers_machines = ts.add_table('gameservers-machines')
    gameservers_machines.set_subfolder_name('ue4-gameservers')
    gameservers_machines.add_primary_key('product_name,group_name,region')
    gameservers_machines.add_foreign_key('product_name', 'products')
    gameservers_machines.add_schema({
        'type': 'object',
        'properties':
        {
            'region': {'type': 'string'},
            'platform': {'enum': ['windows', 'linux']},
            'autoscaling': {'type': 'object', 'properties': {
                'min': {'type': 'integer'},
                'max': {'type': 'integer'},
                'desired': {'type': 'integer'},
                'instance_type': {'type': 'string'},
            }},
        },
    })

    '''
    gameservers-instances:
        gameserver_instance_id  string, pk, default=@@identity
        product_name            string, fk->products, required
        group_name              string, fk->gameservers-machines, required
        region                  string, fk->gameservers-machines, required
        tenant_name             string, fk->tenants, required
        ref                     string, required
        processes_per_machine   integer, required
        command_line            string
    '''
    gameservers_instances = ts.add_table('gameservers-instances')
    gameservers_instances.set_subfolder_name('ue4-gameservers')
    gameservers_instances.add_primary_key('gameserver_instance_id')
    gameservers_instances.add_foreign_key('product_name', 'products')
    # gameservers_instances.add_foreign_key('group_name,region', 'gameservers-machines')
    gameservers_instances.add_foreign_key('tenant_name', 'tenant-names')
    gameservers_instances.add_schema({
        'type': 'object',
        'properties':
        {
            'ref': {'type': 'string'},
            'processes_per_machine': {'type': 'integer'},
            'command_line': {'type': 'string'},
        },
        'required': ['product_name', 'group_name', 'region', 'tenant_name', 'ref', 'processes_per_machine'],
    })
    gameservers_instances.add_default_values({'gameserver_instance_id': '@@identity'})

    '''
    metrics:
        tier_name               string, pk, fk->tiers
        deployable_name         string, pk, fk->deployables
    '''
    metrics = ts.add_table('metrics')
    metrics.add_primary_key('tenant_name')
    metrics.add_foreign_key('tenant_name', 'tenant-names')
    metrics.add_schema({
        'type': 'object',
        'properties':
        {
            's3_bucket': {'type': 'string'},
        },
        'required': ['s3_bucket'],
    })

    # END OF TABLE DEFS

    definition = ts.get_definition()
    new_ts = TableStore()
    new_ts.init_from_definition(definition)
    return new_ts


def push_to_origin(local_ts, force=False, _first=False, _origin_crc=None):
    """
    Pushed 'local_ts' to origin.
    Returns a dict with 'pushed' as True or False depending on success.

    If local store has not been modified since last pull, and the origin
    has the same version, no upload is actually performed and the return
    value contains 'reason' = 'push_skipped_crc_match'.

    If local store has indeed been modified since last pull, but the origin
    has stayed unchanged, upload is performed and the return value contains
    'reason' = 'pushed_to_origin'.

    If origin has changed since last pull, the push is cancelled and the
    return value contains 'reason' = 'checksum_differ'.

    To force a push to a modified origin, set 'force' = True.

    If 'skip_cache' is true, the cache, if defined for the table store, will
    not be updated.

    '_first' is used internally, and indicates it's the first time the table
    store is pushed.

    '_origin_crc' is the original and expected origin crc.
    """
    origin = local_ts.get_table('domain')['origin']
    origin_backend = create_backend(origin)
    origin_ts = None

    if _first:
        crc_match = force = True
    else:
        try:
            origin_ts = origin_backend.load_table_store()
        except Exception as e:
            log.warning("Can't load table store from %s: %s", origin_backend, repr(e))
            crc_match = force = True
        else:
            expected_crc = _origin_crc or local_ts.meta['checksum']
            crc_match = expected_crc == origin_ts.meta['checksum']

    if not force and not crc_match:
        return {
            'pushed': False,
            'reason': 'checksum_differ',
            'local_meta': local_ts.meta.get(),
            'origin_meta': origin_ts.meta.get(),
            'expected_crc': expected_crc,
            # 'local_ts': local_ts,
            # 'origin_ts': origin_ts,
        }

    old, new = local_ts.refresh_metadata()

    if origin_ts is not None:
        crc_match = local_ts.meta['checksum'] == origin_ts.meta['checksum']

    if crc_match and old == new and not force:
        return {'pushed': True, 'reason': 'push_skipped_crc_match'}

    # Always turn on all integrity check when saving to origin
    tmp = driftconfig.relib.CHECK_INTEGRITY
    driftconfig.relib.CHECK_INTEGRITY = ['pk', 'fk', 'unique', 'schema', 'constraints']
    try:
        origin_backend.save_table_store(local_ts)
    finally:
        driftconfig.relib.CHECK_INTEGRITY = tmp

    return {'pushed': True, 'reason': 'pushed_to_origin'}


def pull_from_origin(local_ts, ignore_if_modified=False, force=False):
    origin = local_ts.get_table('domain')['origin']
    origin_ts = create_backend(origin).load_table_store()
    old, new = local_ts.refresh_metadata()

    if old != new and not ignore_if_modified:
        return {'pulled': False, 'reason': 'local_is_modified'}

    crc_match = local_ts.meta['checksum'] == origin_ts.meta['checksum']
    if crc_match and not force:
        return {'pulled': True, 'table_store': local_ts, 'reason': 'pull_skipped_crc_match'}

    return {'pulled': True, 'table_store': origin_ts, 'reason': 'pulled_from_origin'}


def get_redis_cache_backend(ts, tier_name):
    """Returns cache backend for tier 'tier_name' in 'ts'."""
    # Note: drift.core.resources.redis module defines where to find default
    # connection information for a Redis server. We make good use of that here,
    # but it does mean that this piece of code below is now coupled with
    # aforementioned module.
    # Returns None if no cache is defined for the tier and none can be assumed from
    # redis config.
    # TODO: Fix bad coupling as described above.
    domain = ts.get_table('domain').get()
    tier = ts.get_table('tiers').get({'tier_name': tier_name})
    if 'cache' in tier:
        b = create_backend(tier['cache'])
    else:
        if 'resources' not in tier or 'drift.core.resources.redis' not in tier['resources']:
            return None
        redis_info = tier['resources']['drift.core.resources.redis']
        b = RedisBackend.create_from_server_info(
            host=redis_info['host'],
            port=redis_info['port'],
            domain_name=domain['domain_name'],
        )

    return b


def update_cache(ts, tier_name):
    """Push table store 'ts' to its designated Redis cache on tier 'tier_name'."""

    b = get_redis_cache_backend(ts, tier_name)
    if b:
        b.save_table_store(ts)
    return b


class TSTransactionError(RuntimeError):
    pass


class TSTransaction(object):
    _semaphore = 0

    def __init__(self, commit_to_origin=True, write_to_scratch=True):
        self._commit_to_origin = commit_to_origin
        self._write_to_scratch = write_to_scratch
        self._ts = None

    def __enter__(self):
        if self._semaphore != 0:
            raise RuntimeError("Can't nest TSTransactions")

        self._semaphore = 1
        ts, self._url = get_default_drift_config_and_source()
        result = pull_from_origin(ts)
        if not result['pulled']:
            e = TSTransactionError("Can't pull latest table store: {}".format(result['reason']))
            e.result = result
            raise e
        self._ts = result['table_store']
        self._origin_crc = self._ts.meta['checksum']
        self._ts._lock_meta = True

        return self._ts

    def __exit__(self, exc, value, traceback):
        self._semaphore = 0
        self._ts._lock_meta = False

        if exc:
            return False

        if self._commit_to_origin:
            result = push_to_origin(self._ts, _origin_crc=self._origin_crc)
            if not result['pushed']:
                e = TSTransactionError("Can't push to origin: {}".format(result))
                e.result = result
                raise e

        if self._write_to_scratch:
            # Update cache if applicable
            source_backend = create_backend(self._url)
            source_backend.save_table_store(self._ts)


class TSLocal(object):
    def __init__(self):
        self._ts = None

    def __enter__(self):
        self._ts, self._url = get_default_drift_config_and_source()
        return self._ts

    def __exit__(self, exc, value, traceback):
        if exc:
            return False

        # Update cache if applicable
        source_backend = create_backend(self._url)
        source_backend.save_table_store(self._ts)


def parse_8601(s):
    return datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")


def _(d):
    import json
    return json.dumps(d, indent=4, sort_keys=True)


if __name__ == '__main__':
    logging.basicConfig(level='INFO')

    # test TSTransaction

    with TSTransaction('file://~/.drift/config/dgnorth') as ts:
        ts.get_table('domain').get()['display_name'] += ' bluuu!'



# -*- coding: utf-8 -*-
'''

Schema Util - Helpers to pretty print Json schema errors and other bits.

'''
import logging

import jsonschema
from json import dumps
from six import StringIO
from click import echo

log = logging.getLogger(__name__)


def check_schema(json_object, schema, title=None):
    """Do json schema check on object and abort with 400 error if it fails."""
    try:
        jsonschema.validate(json_object, schema, format_checker=jsonschema.FormatChecker())
        # jsonschema.validate(json_object, schema)
    except jsonschema.ValidationError as e:
        report = _generate_validation_error_report(e, json_object)
        if title:
            report = "Schema check failed: %s\n%s" % (title, report)
        e.message = report
        raise


def _generate_validation_error_report(e, json_object):
    """Generate a detailed report of a schema validation error."""

    # Discovering the location of the validation error is not so straight
    # forward:
    # 1. Traverse the json object using the 'path' in the validation exception
    #    and replace the offending value with a special marker.
    # 2. Pretty-print the json object indendented json text.
    # 3. Search for the special marker in the json text to 3 the actual
    #    line number of the error.
    # 4. Make a report by showing the error line with a context of
    #   'lines_before' and 'lines_after' number of lines on each side.

    if json_object is None:
        return "Request requires a JSON body"
    if not e.path:
        return str(e)
    marker = "3fb539de-ef7c-4e29-91f2-65c0a982f5ea"
    lines_before = 7
    lines_after = 7

    # Find the error object and replace it with the marker
    o = json_object
    for entry in list(e.path)[:-1]:
        o = o[entry]
    try:
        orig, o[e.path[0]] = o[e.path[0]], marker
    except Exception:
        # TODO: report the error
        echo("Error setting marker in schemachecker!")

    # Pretty print the object and search for the marker
    json_error = dumps(json_object, indent=4)
    io = StringIO(json_error)

    errline = None
    for lineno, text in enumerate(io):
        if marker in text:
            errline = lineno
            break

    if errline is not None:
        # re-create report
        report = []
        json_object[e.path[0]] = orig
        json_error = dumps(json_object, indent=4)
        io = StringIO(json_error)

        for lineno, text in enumerate(io):
            if lineno == errline:
                line_text = "{:4}: >>>".format(lineno + 1)
            else:
                line_text = "{:4}:    ".format(lineno + 1)
            report.append(line_text + text.rstrip("\n"))

        report = report[
            max(0, errline - lines_before):errline + 1 + lines_after]

        s = "Error in line {}:\n".format(errline + 1)
        s += "\n".join(report)
    else:
        s = str(e)

    return s



import os


def read_version():
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, "VERSION"), "r") as version_file:
        version = version_file.readline().strip()
        return version


__version__ = read_version()

from . import backends  # noqa: F401



# -*- coding: utf-8 -*-
import os
import os.path
import sys
from datetime import datetime, timedelta
import time
import json
import logging
import subprocess
import textwrap
import collections

import click
from click import echo, secho
import six
from six.moves import input

# pygments is optional for now, used for syntax highlighting
try:
    got_pygments = True
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import get_formatter_by_name
except ImportError:
    got_pygments = False

from driftconfig.relib import create_backend, get_store_from_url, diff_meta, diff_tables, CHECK_INTEGRITY, copy_table_store
from driftconfig.config import get_drift_table_store, get_redis_cache_backend, push_to_origin, pull_from_origin, TSTransaction, TSLocal
from driftconfig.config import update_cache
from driftconfig.backends import FileBackend
from driftconfig.util import (
    config_dir, get_domains, get_default_drift_config, get_default_drift_config_and_source,
    define_tenant, prepare_tenant_name, provision_tenant_resources,
    get_tier_resource_modules, register_tier_defaults, register_this_deployable_on_tier,
    register_this_deployable
)
from driftconfig import testhelpers

log = logging.getLogger(__name__)


# Enable simple in-line color and styling of output
try:
    from colorama.ansi import Fore, Back, Style
    styles = {'f': Fore, 'b': Back, 's': Style}
    # Example: "{s.BRIGHT}Bold and {f.RED}red{f.RESET}{s.NORMAL}".format(**styles)
except ImportError:
    class EmptyString(object):
        def __getattr__(self, name):
            return ''

    styles = {'f': EmptyString(), 'b': EmptyString(), 's': EmptyString()}


def get_options(parser):

    subparsers = parser.add_subparsers(
        title="Config file management",
        description="These sets of commands help you with setting up configuration for Drift products.",
        dest="command",
    )

    # 'init' command
    p = subparsers.add_parser(
        'init',
        help='Initialize configuration from a given source.',
        description="Initialize configuration using a given source, like S3, and write it somewhere else, like on local disk.\n"
                    "An example of S3 source: s3://bucket-name/path-name"
    )
    p.add_argument(
        'source',
        action='store',
    )
    p.add_argument(
        '--ignore-errors', '-i',
        action='store_true',
        help='Ignore any errors.'
    )

    # 'list' command
    p = subparsers.add_parser(
        'list',
        help='List locally stored configurations.',
        description="List out all configuration that are stored locally."
    )

    # 'pull' command
    p = subparsers.add_parser(
        'pull',
        help='Pull config.',
        description="Pull latest configuration from source."
    )
    p.add_argument(
        '--loop',
        action='store_true',
        help='pull config continuously for 1 minute.'
    )
    p.add_argument(
        '--ignore-if-modified', '-i',
        action='store_true',
        help='Force a pull from origin even though local version has been modified.'
    )
    p.add_argument(
        '-f', '--force',
        action='store_true',
        help='Force a pull from origin even though local version matches.'
    )
    p.add_argument(
        'domain',
        action='store', nargs='?',
    )

    # 'cache' command
    p = subparsers.add_parser(
        'cache',
        help='Update cache for config.',
        description="Add the config to Redis cache ."
    )
    p.add_argument(
        'domain',
        action='store', nargs='?',
    )
    p.add_argument(
        '-t', '--tier',
        help="The tier on which to update cache, or all if ommitted."
    )

    # 'migrate' command
    p = subparsers.add_parser(
        'migrate',
        help='Migrate config.',
        description="Migrate config to latest definition of TableStore."
    )
    p.add_argument(
        'domain',
        action='store',
    )

    # 'push' command
    p = subparsers.add_parser(
        'push',
        help='Push config.',
        description="Push local config to source. Use with causion."
    )
    p.add_argument(
        'domain',
        action='store',
    )
    p.add_argument(
        '-f', '--force',
        action='store_true',
        help='Force a push to origin even though origin has changed.'
    )

    # 'copy' command
    p = subparsers.add_parser(
        'copy',
        help='Copy config.',
        description="Copy config from one url to another."
    )
    p.add_argument(
        'source_url',
        action='store', help="Source url, or . for default config url."
    )
    p.add_argument(
        'dest_url',
        action='store',
    )
    p.add_argument(
        '-p', '--pickle',
        action='store_true', help="Use pickle format for destination."
    )

    # 'diff' command
    p = subparsers.add_parser(
        'diff',
        help='Diff origin.',
        description="Diff local config to origin."
    )
    p.add_argument(
        'domain',
        action='store', help="Short name to identify the domain or owner of the config.",
    )
    p.add_argument(
        '-d', '--details',
        action='store_true',
        help='Do a detailed diff on modified tables.'
    )

    # MIGRATED FROM drift-admin register command
    p = subparsers.add_parser(
        'register',
        help='Register Drift deployable.',
        description=""
    )
    p.add_argument(
        'project-dir',
        action='store',
        help="Path to project root directory. Default is current working directory.",
        nargs='?',
    )
    p.add_argument(
        "--preview", help="Only preview the changes, do not commit to origin.", action="store_true"
    )

    # MIGRATED FROM drift-admin tenant command suite
    # The create command
    p = subparsers.add_parser(
        'create-tenant',
        help="Create a new tenant for a given product.",
        description="Create a new tenant for a given product."
    )
    p.add_argument(
        'tenant-name',
        action='store',
        help="Name of the tenant.",
    )
    p.add_argument(
        'product-name',
        action='store',
        help="Name of the product.",
    )
    p.add_argument(
        'tier-name',
        action='store',
        help="Name of the tier.",
    )
    p.add_argument(
        '--config',
        help="Specify which config source to use. Will override 'DRIFT_CONFIG_URL' environment variable."
    )
    p.add_argument(
        "--preview", help="Only preview the changes, do not commit to origin.", action="store_true"
    )

    # The refresh command
    p = subparsers.add_parser(
        'refresh-tenant',
        help="Refresh tenant.",
        description="Refresh a tenants on a tier."
    )
    p.add_argument(
        'tenant-name',
        action='store',
        help="Name of the tenant.",
        nargs='?',
    )
    p.add_argument(
        '--config',
        help="Specify which config source to use. Will override 'DRIFT_CONFIG_URL' environment variable."
    )
    p.add_argument(
        "--preview", help="Only preview the changes, do not commit to origin.", action="store_true"
    )

    # The provision command
    p = subparsers.add_parser(
        'provision-tenant',
        help="Provision tenant.",
        description="Provision and prepare resources for a tenant."
    )
    p.add_argument(
        'tenant-name',
        action='store',
        help="Name of the tenant.",
    )
    p.add_argument(
        'deployable-name',
        action='store',
        help="Name of the deployable. Specify 'all' to include all deployables.",
    )
    p.add_argument(
        '--config',
        help="Specify which config source to use. Will override 'DRIFT_CONFIG_URL' environment variable."
    )
    p.add_argument(
        "--preview", help="Only preview the changes, do not commit to origin.", action="store_true"
    )

    # The assign-tier command
    p = subparsers.add_parser(
        'assign-tier',
        help="Assign a deployable to a tier.",
    )
    p.add_argument(
        'deployable-name',
        action='store',
        help="Name of the deployable.",
    )
    p.add_argument(
        "--tiers", help="List of tiers to enable the deployable, or all tiers if omitted.",
        nargs='*',
    )
    p.add_argument(
        "--inactive", help="Mark the deployable inactive. By default the deployable will be marked as active.", action="store_true"
    )
    p.add_argument(
        '--config',
        help="Specify which config source to use. Will override 'DRIFT_CONFIG_URL' environment variable."
    )
    p.add_argument(
        "--preview", help="Only preview the changes, do not commit to origin.", action="store_true"
    )

    # CRUD command suite -------------------------------------------
    # 'create' command
    p = subparsers.add_parser(
        'create',
        description='Create a new Drift configuration.',
    )
    p.add_argument(
        'domain',
        action='store', help="Short name to identify the domain or owner of the config.",
    )
    p.add_argument(
        'source',
        action='store', help="The source location of the config, normally an S3 location."
    )
    p.add_argument(
        '--display-name',
        action='store', help="Display name."
    )


def init_command(args):
    echo("Initializing config from {}".format(args.source))
    if args.ignore_errors:
        del CHECK_INTEGRITY[:]
    ts = create_backend(args.source).load_table_store()
    domain_name = ts.get_table('domain')['domain_name']
    echo("Config domain name: {}".format(domain_name))
    local_store = create_backend('file://' + config_dir(domain_name, user_dir=args.user_dir))
    local_store.save_table_store(ts)
    echo("Config stored at: {}".format(local_store))


def _format_domain_info(domain_info):
    domain = domain_info['table_store'].get_table('domain')
    return "{}: \"{}\" at '{}'. Origin: '{}'".format(
        domain['domain_name'], domain['display_name'], domain_info['path'], domain['origin'])


def list_command(args):
    # Enumerate subfolders at drift/config and see what's there
    domains = get_domains(user_dir=args.user_dir)
    if not domains:
        echo("No Drift configuration found at " + config_dir('', user_dir=args.user_dir))
    else:
        for d in domains.values():
            echo(_format_domain_info(d))


def pull_command(args):
    if args.loop:
        pull_config_loop(args)
    else:
        _pull_command(args)


def pull_config_loop(args):
    # TODO replace with click progress bar
    echo("Starting the pull config loop")
    while now() < end_time:
        st = time.time()
        _pull_command(args)
        diff = time.time() - st
        this_sleep_time = max(sleep_time - diff, 0)
        echo("Waiting for %.1f sec" % this_sleep_time)
        time.sleep(this_sleep_time)
    echo("Completed in %.1f sec" % (now() - start_time).total_seconds())


def _pull_command(args):
    for domain_name, domain_info in get_domains(user_dir=args.user_dir).items():
        if args.domain and args.domain != domain_name:
            continue

        result = pull_from_origin(domain_info['table_store'], ignore_if_modified=args.ignore_if_modified, force=args.force)

        if not result['pulled']:
            echo("Pull failed for " + domain_name + ". Reason: " + result['reason'])
            if result['reason'] == 'local_is_modified':
                echo("Use --ignore-if-modified to overwrite local changes.")
            else:
                echo("Use --force to force a pull.")
        else:
            if result['reason'] == 'pulled_from_origin':
                local_backend = create_backend('file://' + domain_info['path'])
                local_backend.save_table_store(result['table_store'])

            echo("Config for {} pulled. Reason: {}".format(domain_name, result['reason']))


def cache_command(args):
    if args.domain:
        os.environ['DRIFT_CONFIG_URL'] = args.domain
    ts = get_default_drift_config()
    echo("Updating cache for '{}' - {}".format(
        ts.get_table('domain')['domain_name'], ts))

    for tier in ts.get_table('tiers').find():
        tier_name = tier['tier_name']
        if args.tier and args.tier.upper() != tier_name:
            continue
        secho("{}: ".format(tier_name), nl=False, bold=True)
        try:
            b = update_cache(ts, tier_name)
        except Exception as e:
            if "Timeout" not in str(e):
                raise
            secho("Updating failed. VPN down? {}".format(e), fg='red', bold=True)
        else:
            if b:
                secho("Cache updated. Url: {}".format(b.get_url()))
            else:
                secho("No Redis resource defined for this tier.", fg='red', bold=True)

    '''
    # bench test:
    def test_redis_config_fetch(count=10):
        import time
        import os
        from driftconfig.util import get_default_drift_config
        os.environ['DRIFT_CONFIG_URL'] = 'redis://redis.devnorth.dg-api.com/?prefix=dgnorth'
        t = time.time()
        for i in range(count):
            ts = get_default_drift_config()
        t = time.time() - t
        avg = t / count
        echo("Average time to fetch config from redis: %.1f ms." % (avg * 1000.0))
    '''


def migrate_command(args):
    echo("Migrating '{}'".format(args.domain))
    path = config_dir(args.domain, user_dir=args.user_dir)
    if not os.path.exists(path):
        echo("Path not found: " + path)
        sys.exit(1)

    ts = get_drift_table_store()

    class PatchBackend(FileBackend):
        def load_data(self, file_name):
            path_name = self.get_filename(file_name)
            if not os.path.exists(path_name):
                # Attempt to create it just-in-time as a table with zero rows
                head, tail = os.path.split(path_name)
                if not os.path.exists(head):
                    os.makedirs(head)
                with open(path_name, 'w') as f:
                    f.write('[]\n')
            return super(PatchBackend, self).load_data(file_name)

    local_store = PatchBackend(path)
    ts._load_from_backend(local_store, skip_definition=True)
    local_store.save_table_store(ts)


def now():
    return datetime.utcnow()


sleep_time = 10
run_time = 50

start_time = now()
end_time = start_time + timedelta(seconds=run_time)


def push_command(args):
    domain_info = get_domains(user_dir=args.user_dir).get(args.domain)
    if not domain_info:
        echo("Can't push '{}'.".format(args.domain))
        sys.exit(1)

    ts = domain_info['table_store']
    origin = ts.get_table('domain')['origin']
    echo("Pushing local config to source " + origin)
    result = push_to_origin(ts, args.force)
    if not result['pushed']:
        echo("Push failed. Reason: " + result['reason'])
        echo("Origin has changed. Use --force to force push.")
        if 'time_diff' in result:
            echo("Time diff " + result['time_diff'])
    else:
        echo("Config pushed. Reason: " + result['reason'])
        local_store = create_backend('file://' + domain_info['path'])
        local_store.save_table_store(ts)


def copy_command(args):
    echo("Copy '%s' to '%s'" % (args.source_url, args.dest_url))
    if args.source_url == '.':
        ts = get_default_drift_config()
    else:
        ts = get_store_from_url(args.source_url)
    b = create_backend(args.dest_url)
    b.default_format = 'pickle' if args.pickle else 'json'
    b.save_table_store(ts)
    echo("Done.")


def create_command(args):

    domain_info = get_domains(user_dir=args.user_dir).get(args.domain)
    if domain_info:
        echo("The domain name '{}' is taken:".format(args.domain))
        echo(_format_domain_info(domain_info))
        sys.exit(1)

    # Force s3 naming convention. The root folder name and domain name must match.
    if args.source.startswith('s3://'):
        # Strip trailing slashes
        if args.source.endswith('/'):
            args.source = args.source[:-1]

        s3_backend = create_backend(args.source)
        target_folder = s3_backend.folder_name.rsplit('/')[-1]
        if target_folder != args.domain:
            echo("Error: For S3 source, the target folder name and domain name must match.")
            echo("Target folder is '{}' but domain name is '{}'".format(target_folder, args.domain))
            sys.exit(1)
    elif args.source.startswith('file://'):
        # Expand user vars
        args.source = args.source.replace('~', os.path.expanduser('~'))

    # Get empty table store for Drift.
    ts = get_drift_table_store()
    ts.get_table('domain').add(
        {'domain_name': args.domain, 'origin': args.source, 'display_name': args.display_name or ''})

    # Save it locally
    domain_folder = config_dir(args.domain, user_dir=args.user_dir)
    local_store = create_backend('file://' + domain_folder)
    local_store.save_table_store(ts)
    echo("New config for '{}' saved to {}.".format(args.domain, domain_folder))
    echo("Pushing to origin...")
    result = push_to_origin(ts, _first=True)
    if not result['pushed']:
        echo("Push failed. Reason: " + result['reason'])
    echo("Done.")


_ENTRY_TO_TABLE_NAME = {
    'tier': 'tiers',
    'deployable': 'deployable-names',
    'organization': 'organizations',
    'product': 'products',
}


def diff_command(args):
    # Get local table store and its meta state
    domain_info = get_domains(user_dir=args.user_dir).get(args.domain)
    if domain_info is None:
        secho("Configuration not found: {}".format(args.domain), fg='red')
        sys.exit(1)
    local_ts = domain_info['table_store']
    local_m1, local_m2 = local_ts.refresh_metadata()

    # Get origin table store meta info
    origin = local_ts.get_table('domain')['origin']
    origin_backend = create_backend(origin)
    origin_ts = origin_backend.load_table_store()
    origin_meta = origin_ts.meta.get()

    local_diff = ("Local store and scratch", local_m1, local_m2, False)
    origin_diff = ("Local and origin", origin_meta, local_m2, args.details)

    for title, m1, m2, details in local_diff, origin_diff:
        diff = diff_meta(m1, m2)
        if diff['identical']:
            echo(title + " is clean.")
        else:
            echo(title + " are different:")
            echo("\tFirst checksum:  " + diff['checksum']['first'][:7])
            echo("\tSecond checksum: " + diff['checksum']['second'][:7])
            if diff['modified_diff']:
                echo("\tTime since pull: " + str(diff['modified_diff']).split('.')[0])

            echo("\tNew tables: " + str(diff['new_tables']))
            echo("\tDeleted tables: " + str(diff['deleted_tables']))
            echo("\tModified tables: " + str(diff['modified_tables']))

            if details:
                # Diff origin
                origin_ts = get_store_from_url(origin)
                for table_name in diff['modified_tables']:
                    t1 = local_ts.get_table(table_name)
                    t2 = origin_ts.get_table(table_name)
                    tablediff = diff_tables(t1, t2)
                    echo("\nTable diff for " + table_name + "\n(first=local, second=origin):")
                    echo(json.dumps(tablediff, indent=4, sort_keys=True))


def register_command(args):
    project_dir = vars(args)['project-dir'] or '.'
    project_dir = os.path.abspath(project_dir)
    register_deployable(project_dir, args.preview)


def _get_package_info(project_dir):
    """
    Returns info from current package.
    """

    _package_classifiers = [
        'name',
        'version',
        'description',
        'long-description',
        'author',
        'author-email',
        'license'
    ]
    classifier = None
    cmd = [sys.executable, 'setup.py'] + ['--' + c for c in _package_classifiers]
    p = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        cwd=project_dir
    )
    out, err = p.communicate()
    if p.returncode != 0:
        if six.PY3 and err:
            err = err.decode("utf-8")
        raise RuntimeError(
            "Can't get '{}' of this deployable. Error: {} - {}".format(classifier, p.returncode, err)
        )

    info = dict(zip(_package_classifiers, out.decode().splitlines()))
    return info


def register_deployable(project_dir=None, preview=False):

    project_dir = project_dir or '.'
    echo("Project Directory: " + project_dir)

    info = _get_package_info(project_dir)
    name = info['name']

    echo("Registering/updating deployable {}:".format(name))
    echo("Package info:")
    echo(pretty(info))
    echo("")

    # TODO: This is perhaps not ideal, or what?
    config_filename = os.path.join(project_dir, 'config', 'config.json')
    config_filename = os.path.expanduser(config_filename)
    log.info("Loading configuration from %s", config_filename)
    with open(config_filename) as f:
        app_config = json.load(f)

    # Make project_dir importable.
    sys.path.insert(0, project_dir)

    with TSTransaction(commit_to_origin=not preview) as ts:

        ret = register_this_deployable(
            ts=ts,
            package_info=info,
            resources=app_config.get("resources", []),
            resource_attributes=app_config.get("resource_attributes", {}),
        )

        orig_row = ret['old_registration']
        row = ret['new_registration']

        if orig_row is None:
            echo("New registration entry added:")
            echo(pretty(row))
        elif orig_row == row:
            echo("Current registration unchanged:")
            echo(pretty(row))
        else:
            echo("Updating current registration info:")
            echo(pretty(row))
            echo("\nPrevious registration info:")
            echo(pretty(orig_row))

    if preview:
        echo("Preview changes only, not committing to origin.")


def create_tenant_command(args):

    tier_name = vars(args)['tier-name']
    if args.config:
        os.environ['DRIFT_CONFIG_URL'] = args.config

    with TSTransaction(commit_to_origin=not args.preview) as ts:
        prep = prepare_tenant_name(
            ts=ts,
            tenant_name=vars(args)['tenant-name'],
            product_name=vars(args)['product-name']
        )
        tenant_name = prep['tenant_name']

        tenant = ts.get_table('tenant-names').get({'tenant_name': tenant_name})
        if tenant:
            if tenant['tier_name'] != tier_name:
                echo("Tenant '{}' is on tier '{}'. Exiting.".format(tenant_name, tenant['tier_name']))
                sys.exit(1)

            echo("Tenant '{}' already exists. Refreshing it for tier '{}'...".format(
                tenant_name, tier_name))

        result = define_tenant(
            ts=ts,
            tenant_name=tenant_name,
            product_name=prep['product']['product_name'],
            tier_name=tier_name
        )

    echo("Tenant '{}' created/refreshed on tier '{}'.".format(tenant_name, tier_name))
    echo(pretty(result))
    if args.preview:
        echo("\nPreview changes only, not committing to origin.")
        sys.exit(0)


def refresh_tenant_command(args):

    tenant_name = vars(args)['tenant-name']
    echo("Refreshing '{}':".format(tenant_name))
    if args.config:
        os.environ['DRIFT_CONFIG_URL'] = args.config

    # Out of convenience, add current dir to sys.path so the local project can
    # be found during imports.
    sys.path.insert(0, '.')

    with TSTransaction(commit_to_origin=not args.preview) as ts:
        tenant_info = ts.get_table('tenant-names').get({'tenant_name': tenant_name})
        if not tenant_info:
            echo("Tenant '{}' not found!".format(tenant_name))
            sys.exit(1)

        result = define_tenant(
            ts=ts,
            tenant_name=tenant_name,
            product_name=tenant_info['product_name'],
            tier_name=tenant_info['tier_name'],
        )

    echo("Result:")
    echo(pretty(result))
    if args.preview:
        echo("\nPreview changes only, not committing to origin.")
        sys.exit(0)


def provision_tenant_command(args):

    tenant_name = vars(args)['tenant-name']
    deployable_name = vars(args)['deployable-name']

    echo("Provisioning '{}' for {}:".format(tenant_name, deployable_name))
    if deployable_name == 'all':
        deployable_name = None

    if args.config:
        os.environ['DRIFT_CONFIG_URL'] = args.config

    # Out of convenience, add current dir to sys.path so the local project can
    # be found during imports.
    sys.path.insert(0, '.')

    with TSTransaction(commit_to_origin=not args.preview) as ts:
        tenant_info = ts.get_table('tenant-names').get({'tenant_name': tenant_name})
        if not tenant_info:
            echo("Tenant '{}' not found!".format(tenant_name))
            sys.exit(1)

        # Refresh for good measure
        define_tenant(
            ts=ts,
            tenant_name=tenant_name,
            product_name=tenant_info['product_name'],
            tier_name=tenant_info['tier_name'],
        )

        report = provision_tenant_resources(
            ts=ts,
            tenant_name=tenant_name,
            deployable_name=deployable_name,
            preview=args.preview
        )

    echo("Result:")
    echo(pretty(report))
    if args.preview:
        echo("\nPreview changes only, not committing to origin.")
        sys.exit(0)


def assign_tier_command(args):
    deployable_name = vars(args)['deployable-name']
    echo("Assigning '{}':".format(deployable_name))

    if args.config:
        os.environ['DRIFT_CONFIG_URL'] = args.config

    # Out of convenience, add current dir to sys.path so the local project can
    # be found during imports.
    sys.path.insert(0, '.')

    is_active = not args.inactive

    with TSTransaction(commit_to_origin=not args.preview) as ts:
        old_ts = copy_table_store(ts)

        names = [d['deployable_name'] for d in ts.get_table('deployable-names').find()]
        if not names:
            echo("No deployable registered. See 'drift-admin register' for more info.")
            sys.exit(1)

        if deployable_name not in names:
            echo("Deployable '{}' not found. Select one of: {}.".format(
                deployable_name,
                ', '.join(names)
            ))
            sys.exit(1)

        if not args.tiers:
            args.tiers = [tier['tier_name'] for tier in ts.get_table('tiers').find()]

        for tier_name in args.tiers:
            echo("Enable deployable on tier {s.BRIGHT}{}{s.NORMAL}:".format(tier_name, **styles))
            tier = ts.get_table('tiers').get({'tier_name': tier_name})
            if not tier:
                echo("{f.RED}Tier '{}' not found! Exiting.".format(tier_name, **styles))
                sys.exit(1)

            ret = register_this_deployable_on_tier(
                ts, tier_name=tier_name, deployable_name=deployable_name)

            if ret['new_registration']['is_active'] != is_active:
                ret['new_registration']['is_active'] = is_active
                echo("Note: Marking this deployable as {} on tier '{}'.".format(
                    "active" if is_active else "inactive", tier_name))

            # For convenience, register resource default values as well. This
            # is idempotent so it's fine to call it periodically.
            resources = get_tier_resource_modules(
                ts=ts, tier_name=tier_name, skip_loading=False, ignore_import_errors=True)

            # See if there is any attribute that needs prompting,
            # Any default parameter from a resource module that is marked as <PLEASE FILL IN> and
            # is not already set in the config, is subject to prompting.
            tier = ts.get_table('tiers').get({'tier_name': tier_name})
            config_resources = tier.get('resources', {})

            for resource in resources:
                for k, v in resource['default_attributes'].items():
                    if v == "<PLEASE FILL IN>":
                        # Let's prompt if and only if the value isn't already set.
                        attributes = config_resources.get(resource['module_name'], {})
                        if k not in attributes or attributes[k] == "<PLEASE FILL IN>":
                            echo("Enter value for {s.BRIGHT}{}.{}{s.NORMAL}: ".format(
                                resource['module_name'], k, **styles), nl=False)
                            resource['default_attributes'][k] = input()

            echo("\nDefault values for resources configured for this tier:")
            echo(pretty(config_resources))

            register_tier_defaults(ts=ts, tier_name=tier_name, resources=resources)

            echo("\nRegistration values for this deployable on this tier:")
            echo(pretty(ret['new_registration']))
            echo("")

        # Display the diff
        _diff_ts(ts, old_ts)

    if args.preview:
        echo("Preview changes only, not committing to origin.")


def _diff_ts(ts1, ts2):
    from driftconfig.relib import diff_meta, diff_tables
    # Get local table store and its meta state

    ts1 = copy_table_store(ts1)
    ts2 = copy_table_store(ts2)
    local_m1, local_m2 = ts1.refresh_metadata()

    # Get origin table store meta info
    origin_meta = ts2.meta.get()

    title = "Local and origin"
    m1, m2 = origin_meta, local_m2
    diff = diff_meta(m1, m2)

    if diff['identical']:
        echo(title + " is clean.")
    else:
        echo(title + " are different:")
        echo("\tFirst checksum:  " + diff['checksum']['first'][:7])
        echo("\tSecond checksum: " + diff['checksum']['second'][:7])
        if diff['modified_diff']:
            echo("\tTime since pull: " + str(diff['modified_diff']).split('.')[0])

        echo("\tNew tables: {}".format(diff['new_tables']))
        echo("\tDeleted tables: {}".format(diff['deleted_tables']))
        echo("\tModified tables: {}".format(diff['modified_tables']))

        try:
            import jsondiff
        except ImportError:
            echo("To get detailed diff do {s.BRIGHT}pip install jsondiff{s.NORMAL}".format(**styles))
        else:
            # Diff origin
            for table_name in diff['modified_tables']:
                t1 = ts1.get_table(table_name)
                t2 = ts2.get_table(table_name)
                tablediff = diff_tables(t1, t2)
                echo("\nTable diff for {s.BRIGHT}{}{s.NORMAL}".format(table_name, **styles))

                for modified_row in tablediff['modified_rows']:
                    d = json.loads(jsondiff.diff(
                        modified_row['second'], modified_row['first'], dump=True)
                    )
                    echo(pretty(d))


def run_command(args):
    fn = globals()["{}_command".format(args.command.replace("-", "_"))]
    fn(args)


def main(as_module=False):
    import argparse
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('--loglevel', default='WARNING')
    parser.add_argument('--nocheck', action='store_true', help="Skip all relational integrity and schema checks.")
    parser.add_argument('--user-dir', action='store_true', help="Choose user directory over site for locally stored configs.")
    get_options(parser)
    args = parser.parse_args()

    if args.loglevel:
        logging.basicConfig(level=args.loglevel)

    if args.nocheck:
        import driftconfig.relib
        del driftconfig.relib.CHECK_INTEGRITY[:]

    if not args.command:
        parser.print_usage()
        sys.exit(1)
    fn = globals()["{}_command".format(args.command.replace("-", "_"))]
    fn(args)


if __name__ == '__main__':
    main(as_module=True)


def _header(ts):
    domain = ts.get_table('domain')
    secho("Drift config DB ", nl=False)
    secho(domain['domain_name'], bold=True, nl=False)
    secho(" at origin ", nl=False)
    secho(domain['origin'], bold=True)


def _epilogue(ts):
    name = ts.get_table('domain')['domain_name']
    secho("Run \"driftconfig diff {} -d\" to see changes. Run \"driftconfig push {}\" to commit them.".format(name, name))


class Globals(object):
    pass


pass_repo = click.make_pass_decorator(Globals)


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
    '--config-url', '-u', envvar='DRIFT_CONFIG_URL', metavar='',
    help="Url to DB origin.")
@click.option(
    '--verbose', '-v', is_flag=True,
    help='Enables verbose mode.')
@click.option(
    '--organization', '-o', is_flag=True,
    help='Specify organization name/short name.')
@click.option(
    '--product', '-p', is_flag=True,
    help='Specify product name.')
@click.version_option('1.0')
@click.pass_context
def cli(ctx, config_url, verbose, organization, product):
    """This command line tool helps you manage and maintain Drift
    Configuration databases.
    """
    ctx.obj = Globals()
    ctx.obj.config_url = config_url
    if config_url:
        os.environ['DRIFT_CONFIG_URL'] = config_url
    ctx.obj.verbose = verbose
    ctx.obj.organization = organization
    ctx.obj.product = product


@cli.group()
def list():
    """List out information in a configuration DB."""


@list.command()
def configs():
    """List out all Drift configuration DB's that are active on this machine."""
    domains = get_domains()
    if not domains:
        secho(
            "No Drift configuration found on this machine. Run 'init' or 'create' command "
            "to remedy.")
    else:
        got_default = False

        for domain_info in domains.values():
            ts = domain_info['table_store']
            domain = ts.get_table('domain')
            is_default = domain['domain_name'] == ts.get_table('domain')['domain_name']
            if is_default:
                secho(domain['domain_name'] + " [DEFAULT]:", bold=True, nl=False)
                got_default = True
            else:
                secho(domain['domain_name'] + ":", bold=True, nl=False)

            secho(" \"{}\"".format(domain['display_name']), fg='green')
            secho("\tOrigin: " + domain['origin'])
            secho("\tLocal: " + domain_info['path'])
            for tier in ts.get_table('tiers').find():
                b = get_redis_cache_backend(ts, tier['tier_name'])
                if b:
                    secho("\tCache [{}]: {}".format(tier['tier_name'], b.get_url()))
            secho("")

        if got_default:
            if 'DRIFT_CONFIG_URL' in os.environ:
                secho("The default config is specified using the 'DRIFT_CONFIG_URL' environment variable.")
            else:
                secho("The config above is the default one as it's the only one cached locally in ~/.drift/config.")
        else:
            secho("Note: There is no default config specified!")


@list.command()
def deployables():
    """Display registration info for deployables."""
    ts = get_default_drift_config()
    echo("List of Drift deployable plugins in ", nl=False)
    _header(ts)
    deployables = ts.get_table('deployable-names')

    secho("Deployables and api routes:\n", bold=True)

    def join_tables(master_table, tables, search_criteria=None, cb=None):
        """
        Joins rows from 'tables' to the rows of 'master_table' and returns them
        as a single sequence.
        'search_criteria' is applied to the 'master_table'.
        """
        result = []
        rows = master_table.find(search_criteria or {})
        for row in rows:
            row = row.copy()
            for table in tables:
                other = table.get(row)
                if other:
                    row.update(other)
            if cb:
                cb(row)
            result.append(row)
        return result

    def cb(row):
        """Generate list of tier names using info from 'deployables' table."""
        crit = {'deployable_name': row['deployable_name'], 'is_active': True}
        tiers = [d['tier_name'] for d in ts.get_table('deployables').find(crit)]
        row['tiers'] = ', '.join(sorted(tiers))

    head = ['deployable_name', 'api', 'requires_api_key', 'tiers', 'display_name']
    rows = join_tables(
        master_table=deployables,
        tables=[ts.get_table('routing'), ts.get_table('deployable-names')],
        cb=cb,
    )
    tabulate(head, rows, indent='  ')


@list.command()
@click.option('-name', '-n', 'tenant_name', type=str, help="Show full info for given tenant.")
def tenants(tenant_name):
    """Display tenant info."""
    conf = get_default_drift_config()
    _header(conf)

    if tenant_name is None:
        tabulate(
            ['organization_name', 'product_name', 'tier_name', 'tenant_name', 'reserved_at', 'reserved_by'],
            conf.get_table('tenant-names').find(),
            indent='  ',
        )
    else:
        tenant = conf.get_table('tenants').find({'tenant_name': tenant_name})
        if not tenant:
            secho("No tenant named {} found.".format(tenant_name), fg='red', bold=True)
            sys.exit(1)

        secho("Tenant {s.BRIGHT}{}{s.NORMAL}:".format(tenant_name, **styles))
        echo(json.dumps(tenant, indent=4))


@cli.command()
@click.option(
    '--shared', '-s',
    help="Use a shared configuration. The DB will be stored in user directory for your platform, "
    "typically ~/ (or %USERPROFILE% on Windows). The configuration can thus be shared between "
    "multiple projects.",
    is_flag=True
)
@click.option(
    '--recreate',
    help="Recreate Configuration DB. (Note, it will overwrite existing developer config).",
    is_flag=True
)
@click.option(
    '--run', '-r',
    help="Run a Flask server.",
    is_flag=True
)
def developer(recreate, shared, run):
    """Set up environment for local development.

    Creates or updates a Drift Configuration DB. The current deployable is automatically
    registered.
    """
    logging.basicConfig(level='INFO')
    secho("----------------", bold=True)
    secho("Drift Developer:", bold=True)
    secho("----------------", bold=True)

    domain_name = 'developer'
    tier_name = 'LOCALTIER'
    tenant_name = 'developer' if shared else 'localdev'
    origin_folder = '.driftconfig-' + domain_name

    # This here makes sure no external resources are referenced, like AWS services.
    os.environ['DRIFT_USE_LOCAL_SERVERS'] = '1'

    if shared:
        origin_folder = os.path.join(os.path.expanduser('~'), origin_folder)
    else:
        # Make sure origin folder is ignored in git.
        if os.path.exists('.gitignore'):
            with open('.gitignore', 'r+') as f:
                if origin_folder not in f.read():
                    secho("Note! Adding {} to .gitignore".format(click.style(origin_folder, bold=True)))
                    f.seek(0, os.SEEK_END)  # Need this on Windows
                    f.write("\n# Drift config folder for local development\n{}\n".format(origin_folder))

    origin_folder = os.path.abspath(origin_folder)
    secho("Origin of this developer config is at: ", nl=False)
    secho(origin_folder, bold=True, fg='blue')
    origin = 'file://{}'.format(origin_folder)

    # See if config already exists
    try:
        ts = get_store_from_url(origin)
    except Exception:
        ts = None

    if recreate or ts is None:
        if ts:
            secho(
                "Warning: Overriding existing configuration because of --recreate flag.",
                fg='yellow'
            )

        testhelpers.DOMAIN_NAME = domain_name
        testhelpers.ORG_NAME = 'localorg'
        testhelpers.TIER_NAME = tier_name
        testhelpers.PROD_NAME = 'localproduct'
        testhelpers.TENANT_NAME = 'dev'
        config_size = {
            'num_org': 1,
            'num_tiers': 1,
            'num_deployables': 0,
            'num_products': 1,
            'num_tenants': 0,
        }

        ts = testhelpers.create_test_domain(
            config_size=config_size,
            resources=None,
            resource_attributes={}
        )
        domain = ts.get_table('domain').get()
        domain['origin'] = origin
        domain['display_name'] = "Configuration for local development"
    else:
        secho("Using existing developer config DB.")

    project_dir = '.'

    # TODO: This is perhaps not ideal, or what?
    config_filename = os.path.join(project_dir, 'config', 'config.json')
    config_filename = os.path.expanduser(config_filename)
    if not os.path.exists(config_filename) or not os.path.exists('setup.py'):
        secho(
            "Error: Please run this command from a deployable root directory.",
            fg='red', bold=True)
        sys.exit(1)

    log.info("Loading configuration from %s", config_filename)
    with open(config_filename) as f:
        app_config = json.load(f)

    package_info = _get_package_info(project_dir=project_dir)

    sys.path.insert(0, '.')  # Make project_dir importable.

    register_this_deployable(
        ts=ts,
        package_info=package_info,
        resources=app_config.get("resources", []),
        resource_attributes=app_config.get("resource_attributes", {}),
    )
    deployable_name = package_info['name']

    # Make sure this deployable is associated with the current product
    product = ts.get_table('products').find()[0]
    if deployable_name not in product['deployables']:
        product['deployables'].append(deployable_name)

    register_this_deployable_on_tier(
        ts, tier_name='LOCALTIER', deployable_name=deployable_name)

    # For convenience, register resource default values as well. This
    # is idempotent so it's fine to call it periodically.
    resources = get_tier_resource_modules(
        ts=ts, tier_name='LOCALTIER', skip_loading=False, ignore_import_errors=True)

    # See if there is any attribute that needs prompting,
    # Any default parameter from a resource module that is marked as <PLEASE FILL IN> and
    # is not already set in the config, is subject to prompting.
    tier = ts.get_table('tiers').get({'tier_name': 'LOCALTIER'})
    config_resources = tier.get('resources', {})
    USE_EMPTY_AS_DEFAULT = True  # Just to slide through without interruptions.

    for resource in resources:
        for k, v in resource['default_attributes'].items():
            if v == "<PLEASE FILL IN>":
                # Let's prompt if and only if the value isn't already set.
                attributes = config_resources.get(resource['module_name'], {})
                if k not in attributes or attributes[k] == "<PLEASE FILL IN>":
                    attrib_name = '{}.{}'.format(resource['module_name'], k)
                    if attrib_name == 'drift.core.resources.redis.host':
                        value = 'localhost'
                    elif attrib_name == 'drift.core.resources.postgres.server':
                        value = 'localhost'
                    elif USE_EMPTY_AS_DEFAULT:
                        value = ''
                    else:
                        echo("Enter value for {s.BRIGHT}{}{s.NORMAL}: ".format(
                            attrib_name, **styles), nl=False)
                        value = input()
                    resource['default_attributes'][k] = value

    register_tier_defaults(ts=ts, tier_name='LOCALTIER', resources=resources)

    secho("\nDefault values for resources configured for this tier:", bold=True)
    secho(pretty(tier.get('resources', {})))

    # Add the developer tenant
    result = define_tenant(
        ts=ts,
        tenant_name=tenant_name,
        product_name=product['product_name'],
        tier_name=tier_name,
    )

    # Define an alias for the developer tenant so it less cumbersome than the actual name.
    result['tenant_master_row']['alias'] = tenant_name
    tenant_name = result['tenant_master_row']['tenant_name']  # The actual tenant name

    # Provision all resources for the tenant for all deployables
    report = provision_tenant_resources(
        ts=ts,
        tenant_name=tenant_name,
    )

    secho("Provisioned tenant", bold=True)
    secho(pretty(report))

    # Save it locally
    domain_folder = config_dir(domain_name)
    local_store = create_backend('file://' + domain_folder)
    local_store.save_table_store(ts)
    secho("New config for '{}' saved to {}.".format(domain_name, domain_folder))
    result = push_to_origin(ts, _first=True)
    if not result['pushed']:
        secho("Push failed. Reason:" + result['reason'], fg='red')

    secho("\nTo enable local development:\n", bold=True)
    sh = textwrap.dedent("""\
        # The following environment variables define full context for local development
        export DRIFT_CONFIG_URL={}
        export DRIFT_TIER={}
        export DRIFT_DEFAULT_TENANT={}
        export FLASK_APP=drift.devapp:app
        export FLASK_ENV=development

        # To run a flask development server:
        flask run
        """).format(domain_name, tier_name, tenant_name)
    secho(pretty(sh, lexer='bash'))

    # pygments is optionally used for syntax highlightling
    if not got_pygments:
        secho(textwrap.dedent("""\


            Final Note! All the blurb above would look much better with colors!.
            "Plese Run the following command for the sake of rainbows and unicorns:
            "pip install pygments

            """))

    if run:
        # Run a Flask development server
        env = os.environ.copy()
        env['DRIFT_CONFIG_URL'] = domain_name
        env['DRIFT_TIER'] = tier_name
        env['DRIFT_DEFAULT_TENANT'] = tenant_name
        env['FLASK_APP'] = 'drift.devapp:app'
        # allow the user to override the environment, otherwise default to a development env.
        env['FLASK_ENV'] = os.environ.get('FLASK_ENV') or 'development'

        args = ['flask', 'run', '--host=0.0.0.0']
        port = app_config.get('PORT')
        if port is not None:
            args.append('--port={}'.format(port))

        secho(

            "*" * 60 + "\n"
            "* Running on http://0.0.0.0:{}/ (Press CTRL+C to quit)  *\n".format(port or 5000) +
            "*" * 60 + "\n",
            bold=True, fg='red')

        p = subprocess.Popen(
            args,
            cwd=project_dir,
            env=env
        )
        out, err = p.communicate()
        if p.returncode != 0:
            if six.PY3 and err:
                err = err.decode("utf-8")
            raise RuntimeError(
                "Flask run failed: {} - {}".format(p.returncode, err)
            )


@cli.command()
@click.argument('table-name')
def edit(table_name):
    """Edit a config table.\n
    TABLE_NAME is one of: domain, organizations, tiers, deployable-names, deployables,
    products, tenant-names, tenants.
    """
    ts, source = get_default_drift_config_and_source()
    table = ts.get_table(table_name)
    backend = create_backend(source)
    path = backend.get_filename(table.get_filename())
    with open(path, 'r') as f:
        text = click.edit(f.read())
    if text:
        secho("Writing changes to " + path)
        with open(path, 'w') as f:
            f.write(text)

        _epilogue(ts)


@cli.group()
@pass_repo
def tier(repo):
    """Manage tier related entries in the configuration database."""


@tier.command("info")
@click.option('--tier-name', '-t', type=str, default=None)
def tier_info(tier_name):
    """Show tier info."""
    conf = get_default_drift_config()
    _header(conf)
    if tier_name is None:
        echo("Tiers:")
        tabulate(['tier_name', 'state', 'is_live'], conf.get_table('tiers').find(), indent='  ')
    else:
        tier = conf.get_table('tiers').find({'tier_name': tier_name})
        if not tier:
            secho("No tier named {} found.".format(tier_name), fg='red', bold=True)
            sys.exit(1)
        tier = tier[0]
        echo("Tier {}:".format(tier['tier_name']))
        echo(pretty(tier))


@tier.command("add")
@click.argument('tier-name', type=str)
@click.option('--is-live/--is-dev', help="Flag tier for 'live' or 'development' purposes. Default is 'live'.")
@click.option('--edit', '-e', help="Use editor to modify the entry.", is_flag=True)
def tier_add(tier_name, is_live, edit):
    """Add a new tier.\n
    TIER_NAME is a 3-20 character long upper case string containing only the letters A-Z."""
    with TSLocal() as ts:
        tiers = ts.get_table('tiers')
        entry = {'tier_name': tier_name, 'is_live': is_live}
        if edit:
            edit = click.edit(json.dumps(entry, indent=4))
            if edit:
                entry = json.loads(edit)
        if tiers.find(entry):
            secho("Tier {} already exists!".format(entry['tier_name']), fg='red', bold=True)
            sys.exit(1)
        tiers.add(entry)

        _epilogue(ts)


@tier.command("edit")
@click.argument('tier-name', type=str)
def tier_edit(tier_name):
    """Edit a tier."""
    with TSLocal() as ts:
        tiers = ts.get_table('tiers')
        entry = tiers.get({'tier_name': tier_name})
        if not entry:
            secho("tier {} not found!".format(tier_name))
            sys.exit(1)

        edit = click.edit(json.dumps(entry, indent=4))
        if edit:
            entry = json.loads(edit)
            tiers.update(entry)


@cli.group()
@pass_repo
def organization(repo):
    """Manage organizations in the configuration database."""


@organization.command("info")
@click.option(
    '--name', '-n', 'organization_name', type=str,
    help="Show full info for given organization. Specify name or short name.")
def org_info(organization_name):
    """Show organization info."""
    conf = get_default_drift_config()
    _header(conf)

    if organization_name is None:
        tabulate(
            ['organization_name', 'short_name', 'state', 'display_name'],
            conf.get_table('organizations').find(),
            indent='  ',
        )
    else:
        org = conf.get_table('organizations').find({'organization_name': organization_name})
        if not org:
            org = conf.get_table('organizations').find({'short_name': organization_name})
        if not org:
            secho("No organization named {} found.".format(organization_name), fg='red', bold=True)
            sys.exit(1)
        org = org[0]
        echo("Organization {}:".format(org['organization_name']))
        echo(json.dumps(org, indent=4))


@organization.command("add")
@click.argument('organization-name', type=str)
@click.argument('short-name', type=str)
@click.option('--display-name', '-d', help="Display name.", type=str)
@click.option('--edit', '-e', help="Use editor to modify the entry.", is_flag=True)
def org_add(organization_name, short_name, display_name, edit):
    """Add a new organization.\n
    ORGANIZATION_NAME is a 2-20 character long string containing only lower case letters and digits.\n
    SHORT_NAME is a 2-20 character long string containing only lower case letters and digits."""
    with TSLocal() as ts:
        organizations = ts.get_table('organizations')
        entry = {
            'organization_name': organization_name,
            'short_name': short_name,
        }
        if display_name:
            entry['display_name'] = display_name

        if edit:
            edit = click.edit(json.dumps(entry, indent=4))
            if edit:
                entry = json.loads(edit)
        if organizations.find(entry):
            secho("Organization {} already exists!".format(entry['organization_name']), fg='red', bold=True)
            sys.exit(1)
        organizations.add(entry)

        _epilogue(ts)


@organization.command("edit")
@click.argument('organization-name', type=str)
def org_edit(organization_name):
    """Edit a organization."""
    with TSLocal() as ts:
        organizations = ts.get_table('organizations')
        entry = organizations.get({'organization_name': organization_name})
        if not entry:
            secho("organization {} not found!".format(organization_name))
            sys.exit(1)

        edit = click.edit(json.dumps(entry, indent=4))
        if edit:
            entry = json.loads(edit)
            organizations.update(entry)


@cli.group()
@pass_repo
def product(repo):
    """Manage products in the configuration database."""


@product.command("info")
@click.option('-name', '-n', 'product_name', type=str, help="Show full info for given product.")
def product_info(product_name):
    """Show product info."""
    conf = get_default_drift_config()
    _header(conf)

    if product_name is None:
        tabulate(
            ['organization_name', 'product_name', 'state', 'deployables'],
            conf.get_table('products').find(),
            indent='  ',
        )
    else:
        product = conf.get_table('products').find({'product_name': product_name})
        if not product:
            secho("No product named {} found.".format(product_name), fg='red', bold=True)
            sys.exit(1)
        product = product[0]
        secho("Product {s.BRIGHT}{}{s.NORMAL}:".format(product['product_name'], **styles))
        echo(json.dumps(product, indent=4))


@product.command("add")
@click.argument('product-name', type=str)
@click.option('--edit', '-e', help="Use editor to modify the entry.", is_flag=True)
def product_add(product_name, edit):
    """Add a new product.\n
    PRODUCT_NAME is a 3-35 character long string containing only lower case letters digits and dashes.
    The product name must be prefixed with the organization short name and a dash.
    """
    if '-' not in product_name:
        secho(
            "Error: The product name must be prefixed with the organization "
            "short name and a dash.", fg='red', bold=True)
        sys.exit(1)

    short_name = product_name.split('-', 1)[0]
    conf = get_default_drift_config()
    org = conf.get_table('organizations').find({'short_name': short_name})
    if not org:
        secho("No organization with short name {} found.".format(short_name), fg='red', bold=True)
        sys.exit(1)

    organization_name = org[0]['organization_name']

    with TSLocal() as ts:
        products = ts.get_table('products')
        entry = {
            'organization_name': organization_name,
            'product_name': product_name
        }

        if edit:
            edit = click.edit(json.dumps(entry, indent=4))
            if edit:
                entry = json.loads(edit)
        if products.find(entry):
            secho("Product {} already exists!".format(entry['product_name']), fg='red', bold=True)
            sys.exit(1)
        products.add(entry)

        _epilogue(ts)


@product.command("edit")
@click.argument('product-name', type=str)
def product_edit(product_name):
    """Edit a product."""
    with TSLocal() as ts:
        products = ts.get_table('products')
        entry = products.get({'product_name': product_name})
        if not entry:
            secho("product {} not found!".format(product_name))
            sys.exit(1)

        edit = click.edit(json.dumps(entry, indent=4))
        if edit:
            entry = json.loads(edit)
            products.update(entry)


@cli.command(help="View contents of tables. Specify 'all' to view all tables.")
@click.argument('table-name', default='')
@click.option('--tier-name', '-t', 'tier_name', type=str, help="Filter tier.")
@click.option('--tenant-name', '-n', 'tenant_name', type=str, help="Filter tenant.")
@click.option('--deployable-name', '-d', 'deployable_name', type=str, help="Filter deployable.")
def view(table_name, tier_name, tenant_name, deployable_name):
    c = get_default_drift_config()
    if not table_name:
        secho("Specify one the following table names to view, or 'all' for all tables.")
        for table in c.tables:
            secho("\t{}".format(table))
    else:
        def filter(table):
            f = {}
            if tier_name and 'tier_name' in table._pk_fields:
                f['tier_name'] = tier_name
            if tenant_name and 'tenant_name' in table._pk_fields:
                f['tenant_name'] = tenant_name
            if deployable_name and 'deployable_name' in table._pk_fields:
                f['deployable_name'] = deployable_name
            return f

        if table_name == 'all':
            d = {table.name: table.find(filter(table)) for table in c.tables.values()}
        else:
            table = c.get_table(table_name)
            d = table.find(filter(table))

        echo(pretty(d))


@cli.command(help="Set a custom value. Specify either key=value or a json doc.")
@click.argument('key-value')
@click.option('--location', '-l', 'location', type=str, help="Location of entry.\n"
    "If not set the custom value is stored in 'domain' table.\n"
    "For 'tiers' table specify <tier name>\n"
    "For 'tenants' table specify <tenant name>\n"
    "For 'deployables' table specify <tier name>.<deployable name>\n"
    "For any other table specify <table name>[.<pk value 1>.<pk value 2>...]"
    )
@click.option('--raw', '-r', is_flag=True, help="Insert value straight into row.")
@click.option('--preview', '-p', is_flag=True, help="Do not commit changes to config.")
def set(key_value, location, raw, preview):
    try:
        kv = json.loads(key_value)
        if not isinstance(kv, dict):
            secho("Specify either key=value or a json doc.", fg='red')
            sys.exit(1)
    except Exception as e:
        if "=" in key_value:
            key, val = key_value.split('=', 1)
            kv = {key: val}
        else:
            secho("Error: {}".format(e), fg='red')
            sys.exit(1)

    with TSTransaction(commit_to_origin=not preview) as ts:
        if location:
            # Is it tiers?
            table = ts.get_table('tiers')
            row = table.get({'tier_name': location})

            if not row:
                # Is it tenants?
                table = ts.get_table('tenants')
                rows = table.find({'tenant_name': location})
                if rows:
                    row = rows[0]  # We exploit the fact that the tenant name is unique here.

            if not row and '.' in location:
                # Is it deployables?
                table = ts.get_table('deployables')
                tier_name, deployable_name = location.split('.', 1)
                row = table.get({'tier_name': tier_name, 'deployable_name': deployable_name})

            if not row:
                # Is it name of a table?
                if '.' in location:
                    location, filters = location.split('.', 1)
                    filters = filters.split('.')
                else:
                    filters = []
                if location in ts.tables:
                    table = ts.get_table(location)
                    if len(table._pk_fields) != len(filters) :
                        secho("Table '{}' primary keys: {}".format(table.name, ', '.join(table._pk_fields)))
                        secho("No primary key specified, exiting...")
                        sys.exit(0)
                    row = table.get(dict(zip(table._pk_fields, filters)))
            if not row:
                secho("Location '{}' not found!".format(location), fg='red')
                secho("Run 'dconf view' to get a list of available tables.")
                sys.exit(1)
        else:
            table = ts.get_table('domain')
            row = table.get()

        if not raw:
            master_row = row
            row = row.setdefault('custom_values', {})
        else:
            master_row = row

        # From https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
        def dict_merge(dct, merge_dct):
            """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
            updating only top-level keys, dict_merge recurses down into dicts nested
            to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
            ``dct``.
            :param dct: dict onto which the merge is executed
            :param merge_dct: dct merged into dct
            :return: None
            """
            for k, v in merge_dct.items():
                if (k in dct and isinstance(dct[k], dict)
                        and isinstance(merge_dct[k], collections.Mapping)):
                    dict_merge(dct[k], merge_dct[k])
                else:
                    dct[k] = merge_dct[k]

        dict_merge(row, kv)
        secho("Adding custom value to table '{}':".format(table.name))
        echo(pretty(master_row))
        if not preview:
            secho("Committing changes to origin...")

    if preview:
        secho("This is only a preview! No changes made.")
    else:
        secho("Changes committed to origin!", fg='green')


def tabulate(headers, rows, indent=None, col_padding=None):
    """Pretty print tabular data."""
    indent = indent or ''
    col_padding = col_padding or 3

    # Calculate max width for each column
    col_size = [[len(h) for h in headers]]  # Width of header cols
    col_size += [[len(str(row.get(h, ''))) for h in headers] for row in rows]  # Width of col in each row
    col_size = [max(col) for col in zip(*col_size)]  # Find the largest

    # Sort rows
    def make_key(row):
        return ":".join([str(row.get(k, '')) for k in headers])

    rows = sorted(rows, key=make_key)

    for row in [headers] + rows:
        echo(indent, nl=False)
        for h, width in zip(headers, col_size):
            if row == headers:
                h = h.replace('_', ' ').title()  # Make header name pretty
                secho(h.ljust(width + col_padding), bold=True, nl=False)
            else:
                fg = 'white' ##fg = 'black' if row.get('active', True) else 'white'
                secho(str(row.get(h, '')).ljust(width + col_padding), nl=False, fg=fg)
        echo()


PRETTY_FORMATTER = 'console256'
PRETTY_STYLE = 'tango'


def pretty(ob, lexer=None):
    """
    Return a pretty console text representation of 'ob'.
    If 'ob' is something else than plain text, specify it in 'lexer'.

    If 'ob' is not string, Json lexer is assumed.

    Command line switches can be used to control highlighting and style.
    """
    if lexer is None:
        if isinstance(ob, six.string_types):
            lexer = 'text'
        else:
            lexer = 'json'

    if lexer == 'json':
        ob = json.dumps(ob, indent=4, sort_keys=True)

    if got_pygments:
        lexerob = get_lexer_by_name(lexer)
        formatter = get_formatter_by_name(PRETTY_FORMATTER, style=PRETTY_STYLE)
        # from pygments.filters import *
        # lexerob.add_filter(VisibleWhitespaceFilter())
        ret = highlight(ob, lexerob, formatter)
    else:
        ret = ob

    return ret.rstrip()



# -*- coding: utf-8 -*-
import os
import six

from driftconfig.config import get_drift_table_store
from driftconfig.util import (
    set_sticky_config,
    register_this_deployable,
    register_this_deployable_on_tier,
    register_tier_defaults,
    get_tier_resource_modules,
    define_tenant,
    provision_tenant_resources,
)
if six.PY2:
    from string import maketrans
else:
    maketrans = str.maketrans

DOMAIN_NAME = 'testdomain'
ORG_NAME = 'acme'
TIER_NAME = 'LOCALTIER'
DEPL_NAME = 'svc'
PROD_NAME = 'prod'
TENANT_NAME = 'test'

# If serial numbers need to be replaced with letters
tr_upper = maketrans('0123456789', 'ABCDEFGHIJ')
tr_lower = maketrans('0123456789', 'abcdefghij')


def _add(fn, ts, name, config_size, count, **kw):
    if count == 1:
        fn(ts, name, config_size, **kw)
    else:
        for i in six.moves.range(count):
            fn(ts, '{}{}'.format(name, i), config_size, **kw)


def create_test_domain(config_size=None, resources=None, resource_attributes=None):
    """
    Creates a pre-populated Drift config to use for testing.
    'config_size' contains how many entries to create for each tables. The primary keys
    will have a serial number added at the end. See code for more info.
    'resources' is a list of resource module names to include in the config.
    Note that the resources will be provisioned as well. Pass in 'resource_attributes'
    as well for resource module initialization values.
    """
    config_size = config_size or {}
    resources = resources or []

    config_size = {
        'num_org': config_size.get('num_org', 1),
        'num_tiers': config_size.get('num_tiers', 1),
        'num_deployables': config_size.get('num_deployables', 1),
        'num_products': config_size.get('num_products', 1),
        'num_tenants': config_size.get('num_tenants', 1),
    }

    ts = get_drift_table_store()
    ts.get_table('domain').add({
        'domain_name': DOMAIN_NAME,
        'display_name': "Unit Test Domain",
        'origin': ''
    })

    # Always assume local servers
    os.environ['DRIFT_USE_LOCAL_SERVERS'] = '1'

    # First add all the tiers
    # Next, add all the deployables and associate with every tier
    # Then, "customes" can be added, organizations, products and tenants.
    _add(add_tier, ts, TIER_NAME, config_size, config_size['num_tiers'])
    _add(
        add_deployable, ts, DEPL_NAME, config_size, config_size['num_deployables'],
        resources=resources, resource_attributes=resource_attributes,
    )
    _add(add_organization, ts, ORG_NAME, config_size, config_size['num_org'])
    set_sticky_config(ts)

    if resources:
        for tenant in ts.get_table('tenant-names').find():
            provision_tenant_resources(ts=ts, tenant_name=tenant['tenant_name'])

    return ts


def terminate_tenants(ts):
    """Terminate (or unprovision) all tenants."""
    tenants = ts.get_table('tenants')
    for tenant in ts.get_table('tenant-names').find():
        for tenant_details in tenants.find({'tenant_name': tenant['tenant_name']}):
            tenant_details['state'] = 'uninitializing'
        provision_tenant_resources(ts=ts, tenant_name=tenant['tenant_name'])


def add_tier(ts, tier_name, config_size):

    # Can't have numbers in tier names
    tier_name = tier_name.translate(tr_upper)

    ts.get_table('tiers').add({
        'tier_name': tier_name,
        'is_live': True,
        'aws': {
            'region': 'test-region-1',
            'ssh_key': 'unittest-key',
        },
    })


def add_deployable(ts, deployable_name, config_size, resources, resource_attributes):
    package_info = {'name': deployable_name, 'description': deployable_name}

    register_this_deployable(
        ts=ts,
        package_info=package_info,
        resources=resources,
        resource_attributes=resource_attributes,
    )

    for tier in ts.get_table('tiers').find():
        register_this_deployable_on_tier(
            ts=ts,
            tier_name=tier['tier_name'],
            deployable_name=deployable_name,
        )

        resources = get_tier_resource_modules(
                ts=ts, tier_name=tier['tier_name'], skip_loading=False)
        register_tier_defaults(ts=ts, tier_name=tier['tier_name'], resources=resources)


def add_organization(ts, organization_name, config_size):
    ts.get_table('organizations').add({
        'organization_name': organization_name,
        'short_name': organization_name,
        'display_name': 'Some Test Organization',
    })

    # Each organization has x many products
    _add(
        add_product, ts, PROD_NAME, config_size, config_size['num_products'],
        organization_name=organization_name,
    )


def add_product(ts, product_name, config_size, organization_name):

    product_name = '{}-{}'.format(organization_name, product_name)
    deployables = [d['deployable_name'] for d in ts.get_table('deployables').find()]

    ts.get_table('products').add({
        'product_name': product_name,
        'organization_name': organization_name,
        'deployables': deployables,
    })

    # Each product has x many tenants
    _add(
        add_tenant, ts, TENANT_NAME, config_size, config_size['num_tenants'],
        product_name=product_name, organization_name=organization_name
    )


def add_tenant(ts, tenant_name_prefix, config_size, product_name, organization_name):

    for tier in ts.get_table('tiers').find():
        tenant_name = '{}-{}'.format(product_name, tenant_name_prefix)
        tier_name = tier['tier_name']
        if config_size['num_tiers'] != 1:
            tenant_name += '-{}'.format(tier_name.lower())

        # Define a tenant
        define_tenant(
            ts, tenant_name=tenant_name, product_name=product_name, tier_name=tier_name)


def get_name(which):
    """
    Returns the generated name of a tier, product, org, or tenant.
    """
    if which == 'organization':
        return ORG_NAME
    elif which == 'tier':
        return TIER_NAME
    elif which == 'deployable':
        return DEPL_NAME
    elif which == 'product':
        return '{}-{}'.format(ORG_NAME, PROD_NAME)
    elif which == 'tenant':
        return '{}-{}'.format(get_name('product'), TENANT_NAME)



# -*- coding: utf-8 -*-
import logging
from datetime import datetime
import collections
import sys
import os
import os.path
import getpass
import importlib
from cachetools import cached, TTLCache

from click import echo

from driftconfig.relib import get_store_from_url, create_backend

log = logging.getLogger(__name__)


config_source_cache = TTLCache(maxsize=10, ttl=3.0)
cache_hits = 0
cache_misses =0


def set_cache_ttl(ttl):
    """Set a different timeout for config source cache."""
    global config_source_cache
    config_source_cache.clear()
    config_source_cache = TTLCache(maxsize=10, ttl=ttl)


class ConfigNotFound(RuntimeError):
    """Raised if no config or multiple configs are found."""
    pass


class TenantNotConfigured(RuntimeError):
    """Raised in case a tenant is not found during a call to get_drift_config()."""
    pass


def config_dir(config_name, user_dir=False):
    """
    Return a path to config dir for 'config_name'.
    If 'user_dir' is set, the path is in the user directory, else it's in
    shared directory for all users.
    """
    # HACK: Use user dir always, except for windows
    if sys.platform == 'win32':
        user_dir = False
    else:
        user_dir = True

    if user_dir:
        return os.path.join(os.path.expanduser('~'), '.drift', 'config', config_name)
    else:
        if sys.platform == 'win32':
            root = os.environ.get('ProgramData')
        elif sys.platform == 'darwin':
            root = '/Library/Application Support'
        else:
            root = '/etc'

        return os.path.join(root, 'drift', 'config', config_name)


def get_domains(user_dir=False, skip_errors=False):
    """Return all config domains stored on local disk."""
    config_folder = config_dir('', user_dir=user_dir)
    domains = {}
    if not os.path.exists(config_folder):
        return {}

    for dir_name in os.listdir(config_folder):
        path = os.path.join(config_folder, dir_name)
        if os.path.isdir(path):
            try:
                ts = get_store_from_url('file://' + path)
            except Exception as e:
                if skip_errors:
                    log.warning("Note: '%s' is not a config folder or is corrupt. (%s).", path, e)
                    continue
                else:
                    raise
            domain = ts.get_table('domain')
            domains[domain['domain_name']] = {'path': path, 'table_store': ts}
    return domains


_sticky_ts = None


def set_sticky_config(ts):
    """Assign permanently 'ts' as the one and only drift config. Useful for tests."""
    global _sticky_ts
    _sticky_ts = ts


def get_default_drift_config():
    """
    Return Drift config as a table store.
    If 'DRIFT_CONFIG_URL' is found in environment variables, it is used to load the
    table store. If not found, the local disk is searched using get_domain() and if
    only one configuration is found there, it is used.
    If all else fails, this function raises an exception.

    If a table store object was set using set_sticky_config(), then that object will
    always be returned.
    """
    ts, source = get_default_drift_config_and_source()
    return ts


def get_default_drift_config_and_source():
    """
    Same as get_default_drift_config but returns a tuple of table store and the
    source of where it was loaded from.
    """
    global cache_hits
    global cache_misses

    if _sticky_ts:
        return _sticky_ts, 'memory://_dummy'

    url = os.environ.get('DRIFT_CONFIG_URL')
    if url:
        ts = config_source_cache.get(url)
        if ts:
            cache_hits += 1
            return ts, url

        # Enable domain shorthand
        if ':' not in url:
            domains = get_domains()
            domain = domains.get(url)
            if domain:
                return domain['table_store'], 'file://' + domain['path']
            else:
                raise RuntimeError("No domain named '{}' found on local disk. Available domains: {}.".format(
                    url, ", ".join(domains.keys())))

        b = create_backend(url)
        ts = b.load_table_store()
        if b.is_cache:
            cache_misses += 1
            config_source_cache[url] = ts
        return ts, url
    else:
        domains = get_domains()
        if len(domains) == 0:
            raise ConfigNotFound(
                "No config found in ~/.drift/config. Use 'driftconfig init' command to "
                "initialize a local config, or add a reference to the config using the "
                "environment variable 'DRIFT_CONFIG_URL'."
            )
        elif len(domains) != 1:
            domain_names = ", ".join(domains.keys())
            raise ConfigNotFound(
                "Multiple Drift configurations found in ~/.drift/config.\n"
                "Specify which configuration to use by referencing it in the "
                "'DRIFT_CONFIG_URL' environment variable.\n"
                "Configurations available on local disk: %s."
                "" % domain_names)
        domain = next(d for d in domains.values())
        return domain['table_store'], 'file://' + domain['path']


conf_tuple = collections.namedtuple(
    'driftconfig',
    [
        'table_store',
        'organization',
        'product',
        'tenant_name',
        'tier',
        'deployable',
        'tenant',
        'tenants',
        'domain',
        'drift_app',
        'source',
    ]
)


def get_drift_config(
    ts=None,
    tenant_name=None,
    tier_name=None,
    deployable_name=None,
    drift_app=None,
    allow_missing_tenant=False
):
    """
    Return config tuple for given config context, containing the following properties:
    ['organization', 'product', 'tenant_name', 'tier', 'deployable', 'tenant']

    'ts' is the config TableStore object.
    'tenant_name' is the name of the tenant, or None if not applicable.

    If 'tenant_name' is specified but not found in config a TenantNotConfigured exception
    is raised. If 'allow_missing_tenant' is True however, then the config tuple will be
    returned but with the 'tenant' property set to None.
    """
    if ts:
        source = "internal"
    else:
        ts, source = get_default_drift_config_and_source()

    # Map tenant alias to actual tenant name if needed.
    tenant_name_row = ts.get_table('tenant-names').find({'alias': tenant_name})
    alias = tenant_name
    if tenant_name_row:
        tenant_name = tenant_name_row[0]['tenant_name']
        alias = "{} (alias={})".format(tenant_name, alias)

    # HACK BEGIN: Until 'flask_config' has been repurposed into 'drift_app' config, we enable a little mapping between
    # here for convenience:
    if drift_app and not deployable_name:
        deployable_name = drift_app['name']
    # HACK END

    tenants = ts.get_table('tenants')
    if tenant_name:
        tenant = tenants.get({'tier_name': tier_name, 'deployable_name': deployable_name, 'tenant_name': tenant_name})

        if not tenant and not allow_missing_tenant:
            raise TenantNotConfigured(
                "Tenant '{}' not found for tier '{}' and deployable '{}'".format(alias, tier_name, deployable_name)
            )
    else:
        tenant = None

    if tenant:
        tenant_name = tenants.get_foreign_row(tenant, 'tenant-names')
        product = ts.get_table('tenant-names').get_foreign_row(tenant_name, 'products')
        organization = ts.get_table('products').get_foreign_row(product, 'organizations')
    else:
        tenant_name = None
        product = None
        organization = None

    if tier_name and deployable_name:
        tenant_rows = tenants.find(
            {'tier_name': tier_name, 'deployable_name': deployable_name, 'state': 'active'}
        )
    else:
        tenant_rows = []

    conf = conf_tuple(
        table_store=ts,
        tenant=tenant,
        tier=ts.get_table('tiers').get({'tier_name': tier_name}),
        deployable=ts.get_table('deployables').get({'deployable_name': deployable_name, 'tier_name': tier_name}),
        domain=ts.get_table('domain'),
        tenant_name=tenant_name,
        tenants=tenant_rows,
        product=product,
        organization=organization,
        drift_app=drift_app,
        source=source,
    )

    # Make sure if tier name was specified, that it actually exists.
    if tier_name and conf.tier is None:
        raise RuntimeError(
            "Tier '{}' not found in config '{}'.".format(tier_name, conf.domain['domain_name']))

    return conf


def prepare_tenant_name(ts, tenant_name, product_name):
    """
    Prepares a tenant name by prefixing it with the organization shortname and returns the
    product and organization record associated with it. The value of 'tenant_name' may
    already contain the prefix.
    Returns a dict of 'tenant_name', 'product' and 'organization'.
    """
    products = ts.get_table('products')
    product = products.get({'product_name': product_name})
    if not product:
        raise RuntimeError("Product '{}' not found.".format(product_name))

    organization = products.get_foreign_row(product, 'organizations')

    if '-' in tenant_name:
        org_short_name = tenant_name.split('-', 1)[0]
        if org_short_name != organization['short_name']:
            raise RuntimeError("Tenant name '{}' must be prefixed with '{}'.".format(
                tenant_name, org_short_name)
            )
    else:
        tenant_name = '{}-{}'.format(organization['short_name'], tenant_name)

    return {
        'tenant_name': tenant_name,
        'product': product,
        'organization': organization,
    }


def define_tenant(ts, tenant_name, product_name, tier_name):
    """
    Defines a new tenant or updates/refreshes a current one.
    Table store 'ts' is updated accordingly.

    If 'tenant_name' does not exist, a new tenant record is created in 'tenant-names' table
    and new record is created in 'tenants' for each deployable that is enabled for
    the product identified by 'product_name'.

    If 'tenant_name' exists, and deployables for the product have changed (new one added or
    a current one removed), the 'tenants' table will be updated accordingly. In case of new
    deployable a new record will simply be added. In case a deployable is inactive or removed
    from the given product, the tenant record for the deployable will be put in state 'disabled'
    or 'deleted' respectively.

    Returns same dict as 'prepare_tenant_name' with addition of 'report' which contains a
    list of [deployable_name, state] tuples indicating state of each deployable for this
    tenant after the creation/update command.
    """
    prep = prepare_tenant_name(ts=ts, tenant_name=tenant_name, product_name=product_name)
    tenant_name = prep['tenant_name']
    product = prep['product']
    organization = prep['organization']

    # Add a record to 'tenant-names' if needed.
    tenant_names = ts.get_table('tenant-names')
    tenant_master_row = tenant_names.get({'tenant_name': tenant_name})
    if not tenant_master_row:
        tenant_master_row = tenant_names.add({
            'tenant_name': tenant_name,
            'organization_name': organization['organization_name'],
            'product_name': product_name,
            'tier_name': tier_name,
            'reserved_by': getpass.getuser(),
            'reserved_at': datetime.utcnow().isoformat() + 'Z',
        })

    prep['tenant_master_row'] = tenant_master_row

    # Make a list of active and inactive deployables associated with the given product.
    active_deployables = []
    inactive_deployables = []
    deployables = ts.get_table('deployables')
    for deployable_name in product['deployables']:
        deployable = deployables.get({'tier_name': tier_name, 'deployable_name': deployable_name})
        if deployable:
            if deployable['is_active']:
                active_deployables.append(deployable_name)
            else:
                inactive_deployables.append(deployable_name)
        else:
            raise RuntimeError(
                "Deployable '{}' defined for product '{}' is not found in table 'deployables' for tier {}.".format(
                    deployable_name, product_name, tier_name)
            )

    tenants = ts.get_table('tenants')
    report = []  # List of deployable names and current state.

    def add_report(deployable_name, state):
        report_row = {'deployable_name': deployable_name, 'state': state}
        report.append(report_row)
        return report_row

    # Deactivate/delete deployables if needed
    for tenant in tenants.find({'tier_name': tier_name, 'tenant_name': tenant_name}):
        deployable_name = tenant['deployable_name']
        if deployable_name in inactive_deployables:
            tenant['state'] = 'disabled'
            add_report(deployable_name, 'disabled')
        elif deployable_name not in active_deployables and tenant['state'] != 'deleted':
            tenant['state'] = 'uninitializing'  # Signal de-provision of resources.
            add_report(deployable_name, 'uninitializing')

    # Activate/associate deployables if needed
    for deployable_name in active_deployables:
        pk = {
            'tier_name': tier_name,
            'deployable_name': deployable_name,
            'tenant_name': tenant_name
        }
        tenant = tenants.get(pk)
        if tenant:
            report_row = add_report(deployable_name, tenant['state'])
        else:
            # State is set to 'initializing' by default signaling provisioning of resources.
            tenant = tenants.add(pk)
            report_row = add_report(deployable_name, 'initializing')

        # Initialize/update tenant resource config using default tier or deployables values.
        depl_names = ts.get_table('deployable-names').get({'deployable_name': deployable_name})
        tier = ts.get_table('tiers').get({'tier_name': tier_name})
        for resource_name in depl_names['resources']:
            # LEGACY SUPPORT: Shorten resource name to its last bit
            legacy_resource_name = resource_name.rsplit('.', 1)[1]

            # Update the tenants attributes but leave current ones intact
            resource_attribs = tenant.setdefault(legacy_resource_name, {})

            # Apply tier defaults
            for k, v in tier['resources'][resource_name].items():
                resource_attribs.setdefault(k, v)

            # Apply deployable defaults
            deployable_defaults = depl_names['resource_attributes'].get(resource_name, {})
            for k, v in deployable_defaults.items():
                resource_attribs.setdefault(k, v)

            report_row.setdefault('resources', {})[resource_name] = resource_attribs

    prep['report'] = report
    return prep


def provision_tenant_resources(ts, tenant_name, deployable_name=None, preview=False):
    """
    Calls resource provisioning functions for tenant 'tenant_name'.
    If 'deployable_name' is set, only the resource modules for that deployable
    are called.
    If 'preview' then the provision_resource() callback function is not called.
    """
    tenant_info = ts.get_table('tenant-names').get({'tenant_name': tenant_name})
    termination_protection = tenant_info.get('termination_protection', False)
    crit = {'tenant_name': tenant_name, 'tier_name': tenant_info['tier_name']}
    if deployable_name:
        crit['deployable_name'] = deployable_name
    configurations = ts.get_table('tenants').find(crit)

    report = {
        'tenant': tenant_info,
        'deployables': {},
    }

    log.info("Provisioning tenant '%s'", tenant_name)

    for tenant_config in configurations:
        # LEGACY SUPPORT: Need to look up the actual resource module name
        depl = ts.get_table('deployable-names').get({'deployable_name': tenant_config['deployable_name']})
        depl_report = {
            'resources': {}
        }
        report['deployables'][tenant_config['deployable_name']] = depl_report

        log.info("  Deployable: '%s'", tenant_config['deployable_name'])

        # Termination protection is specified for the tenant as a whole, but we need to see if
        # individual deployable is uninitializing and stop it there.
        if tenant_config['state'] == 'uninitializing' and termination_protection:
            depl_report['error'] = "Tenant has termination protection. Can't uninitialize."
            continue

        for dryrun in [True, False]:
            for resource_module in depl['resources']:
                legacy_resource_name = resource_module.rsplit('.', 1)[1]
                resource_attributes = tenant_config.setdefault(legacy_resource_name, {})

                def report_error():
                    result = [
                        "Failed to provision resource '{}'.\n{}: {}\nAttributes: {}".format(
                            resource_module, e.__class__.__name__, e, resource_attributes)
                        ]
                    depl_report['resources'][resource_module] = result

                try:
                    m = importlib.import_module(resource_module)
                except Exception as e:
                    report_error()
                    continue

                has_provision_method = hasattr(m, 'provision_resource')

                if has_provision_method:
                    log.info("  -> %s", resource_module)

                if dryrun:  # Just verifying that all the resource modules can load properly
                    if hasattr(m, 'provision_resource_precheck'):
                        m.provision_resource_precheck()
                    continue

                if has_provision_method and not preview:
                    try:
                        result = m.provision_resource(
                            ts=ts,
                            tenant_config=tenant_config,
                            attributes=resource_attributes,
                        )
                        depl_report['resources'][resource_module] = result
                    except Exception as e:
                        report_error()
                else:
                    depl_report['resources'][resource_module] = resource_attributes

        depl_report['old_state'] = tenant_config['state']

        if tenant_config['state'] == 'initializing':
            tenant_config['state'] = 'active'
        elif tenant_config['state'] == 'uninitializing':
            tenant_config['state'] = 'deleted'

        depl_report['new_state'] = tenant_config['state']

    return report


def refresh_tenants(ts, tenant_name=None, tier_name=None):
    """
    Refreshes config info for a tenant by calling define_tenant().
    If 'tenant_name' is set, only that tenant is refreshed.
    If 'tier_name' is set, only tenants defined on that tier are refreshed.
    """
    if tier_name:
        tiers = [tier_name]
    else:
        crit = {'tenant_name': tenant_name} if tenant_name else {}
        tiers = set(t['tier_name'] for t in ts.get_table('tenants').find(crit))

    if tenant_name:
        tenant_names = [tenant_name]
    else:
        crit = {'tier_name': tier_name} if tier_name else {}
        tenant_names = set(t['tenant_name'] for t in ts.get_table('tenants').find(crit))

    for tenant in ts.get_table('tenants').find():
        if tenant['tier_name'] not in tiers:
            continue
        if tenant['tenant_name'] not in tenant_names:
            continue

        tenant_info = ts.get_table('tenant-names').get({'tenant_name': tenant['tenant_name']})

        yield define_tenant(
            ts=ts,
            tenant_name=tenant['tenant_name'],
            product_name=tenant_info['product_name'],
            tier_name=tenant['tier_name']
        )


# NOTE THIS IS DEPRECATED FUNCTION AND NEEDS TO BE UPGRADED TO NU STYLE SUMTHIN
def get_parameters(config, args, required_keys, resource_name):
    raise NotImplementedError()
    echo("ROUESR IDC NOAME IS " + resource_name)
    defaults = config.tier.get('resource_defaults', [])

    # gather default parameters from tier
    params = {}
    for default_params in defaults:
        if default_params.get("resource_name") == resource_name:
            params = default_params["parameters"].copy()

    if not params:
        raise RuntimeError("No provisioning defaults in tier config for '%s'. Cannot continue" % resource_name)
    if set(required_keys) != set(params.keys()):
        log.error("%s vs %s" % (required_keys, params.keys()))
        raise RuntimeError("Tier provisioning parameters do not match tier defaults for '%s'. Cannot continue" % resource_name)
    for k in args.keys():
        if k not in params:
            raise RuntimeError("Custom parameter %s for '%s' not supported. Cannot continue" % (k, resource_name))

    log.info("Default parameters for '%s' from tier: %s", resource_name, params)
    params.update(args)
    if args:
        log.info("Connection info for '%s' with custom parameters: %s", resource_name, params)
    return params


def register_this_deployable(ts, package_info, resources, resource_attributes):
    """
    Registers top level information for a deployable package.

    'package_info' is a dict containing Python package info for current deployable. The dict contains
    at a minimum 'name' and 'description' fields.

    'resources' is a list of resource modules in use by this deployable.

    'resource_attributes' is a dict containing optional attributes for resources. Key is the resource
    module name, value is a dict of any key values attributes which gets passed into registration callback
    functions in the resource modules.

    Returns a dict with 'old_registration' row and 'new_registration' row.
    """
    tbl = ts.get_table('deployable-names')
    pk = {'deployable_name': package_info['name']}
    orig_row = tbl.get(pk)
    if orig_row:
        row = orig_row
        orig_row = orig_row.copy()
    else:
        pk['display_name'] = package_info['description']  # This is a required field
        row = tbl.add(pk)

    row['display_name'] = package_info['description']
    if 'long-description' in package_info and package_info['long-description'] != "UNKNOWN":
        row['description'] = package_info['long-description']
    row['resources'] = resources
    row['resource_attributes'] = resource_attributes

    # Call hooks for top level registration.
    for module_name in row['resources']:
        m = importlib.import_module(module_name)
        if hasattr(m, 'register_deployable'):
            attributes = resource_attributes.setdefault(module_name, {})
            m.register_deployable(
                ts=ts,
                deployablename=row,
                attributes=attributes,
            )

            # For cleanliness, remove resource attribute entry if it's empty
            if not attributes:
                del resource_attributes[module_name]

    return {'old_registration': orig_row, 'new_registration': row}


def register_this_deployable_on_tier(ts, tier_name, deployable_name):
    """
    Registers tier specific info for a deployable package.

    Returns a dict with 'old_registration' row and 'new_registration' row.
    """
    tbl = ts.get_table('deployables')
    pk = {'tier_name': tier_name, 'deployable_name': deployable_name}
    orig_row = tbl.get(pk)
    if orig_row:
        row, orig_row = orig_row, orig_row.copy()
    else:
        row = tbl.add(pk)

    registration_row = ts.get_table('deployable-names').get({'deployable_name': deployable_name})
    resource_attributes = registration_row['resource_attributes']

    # Call hooks for tier registration info.
    for module_name in registration_row['resources']:
        m = importlib.import_module(module_name)
        if hasattr(m, 'register_deployable_on_tier'):
            attributes = resource_attributes.setdefault(module_name, {})
            m.register_deployable_on_tier(
                ts=ts,
                deployable=row,
                attributes=attributes,
            )

            # For cleanliness, remove resource attribute entry if it's empty
            if not attributes:
                del resource_attributes[module_name]

    return {'old_registration': orig_row, 'new_registration': row}


def get_tier_resource_modules(ts, tier_name, skip_loading=False, ignore_import_errors=False):
    """
    Returns a list of all resource modules registered on 'tier_name'.
    Each entry is a dict with 'module_name', 'module' and 'default_attributes'.
    If 'skip_loading' the 'module' value is None.
    If 'ignore_import_errors' import errors will be ignored.
    """
    resources = set()
    deployables = ts.get_table('deployables')
    for deployable in deployables.find({'tier_name': tier_name}):
        row = deployables.get_foreign_row(deployable, 'deployable-names')
        resources |= set(row.get('resources', []))  # Update resource module list

    modules = []
    for module_name in resources:
        if skip_loading:
            m = None
        else:
            try:
                m = importlib.import_module(module_name)
            except ImportError as e:
                if not ignore_import_errors:
                    raise
                log.warning("Ignoring import error: %s", e)
                continue
        modules.append({
            'module_name': module_name,
            'module': m,
            'default_attributes': getattr(m, 'TIER_DEFAULTS', {}),
        })

    return modules


def register_tier_defaults(ts, tier_name, resources=None):
    """
    Registers tier specific default values for resources.
    Note, this is deployable-agnostic info.
    If 'resources' is set, it must originate from get_tier_resource_modules(). This is to
    give the option of modifying values in 'default_attributes'.
    """
    # Enumerate all resource modules from all registered deployables on this tier,
    # configure default vaules and call hooks for tier registration info.
    tier = ts.get_table('tiers').get({'tier_name': tier_name})
    if resources is None:
        module_resources = get_tier_resource_modules(ts=ts, tier_name=tier_name)
    else:
        module_resources = resources

    config_resources = tier.setdefault('resources', {})

    for resource in module_resources:
        # Create or refresh default attributes entry in config
        attributes = config_resources.setdefault(resource['module_name'], {})
        # Only add new entries to 'attributes' otherwise we override some values with placeholder data.
        for k, v in resource['default_attributes'].items():
            if k not in attributes or attributes[k] == "<PLEASE FILL IN>":
                attributes[k] = v

        # Give resource module chance to do custom work
        if hasattr(resource['module'], 'register_resource_on_tier'):
            resource['module'].register_resource_on_tier(
                ts=ts,
                tier=tier,
                attributes=attributes,
            )
